"""
Integration testing tools for processes with blocking event loops
"""
import asyncio, uvloop
import zmq.asyncio
import random
import inspect
import traceback
import time
import os
from cilantro_ee.utils.lprocess import LProcess
from cilantro_ee.core.crypto import wallet
from cilantro_ee.core.logger import get_logger, overwrite_logger_level
from vmnet.testcase import BaseNetworkTestCase

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


ASSERTS_POLL_FREQ = 0.1
CHILD_PROC_TIMEOUT = 1

TEST_PROC_SETUP_TIME = 20

MPTEST_PORT = '10200'

SIG_RDY = b'IM RDY'
SIG_SUCC = b'G00DSUCC'
SIG_FAIL = b'BADSUCC'
SIG_ABORT = b'NOSUCC'
SIG_START = b'STARTSUCC'


def _run_test_proc(name, url, build_fn, config_fn, assert_fn, log_lvl=0, env_vars={}):
    log = get_logger("TestObjectRunner[{}]".format(name))

    for k, v in env_vars.items():
        log.debug("Setting env var {}={}".format(k, v))
        os.environ[k] = v

    if log_lvl:
        log.notice("Setting log level to {}".format(log_lvl))
        overwrite_logger_level(log_lvl)

    # TODO create socket outside of loop and pass it in for
    tester = MPTesterProcess(name=name, url=url, build_fn=build_fn, config_fn=config_fn, assert_fn=assert_fn)
    log.debug("MPTesterProcess named {} created".format(name))
    tester_socket = tester.socket

    try:
        tester.start_test()
    except Exception as e:
        log.error("\n\n TesterProcess encountered exception outside of internal loop! Error:\n {}\n\n"
                       .format(traceback.format_exc()))
        tester_socket.send(SIG_FAIL)
        tester._teardown()


def wrap_fn(func, *args, **kwargs):
    def _func():
        return func(*args, **kwargs)
    return _func


def get_filename(proc_name):
    import os
    prefix = os.getenv('HOST_NAME', '')
    if prefix:
        proc_name = prefix + '_' + proc_name
    return "{}_{}".format(proc_name, os.getpid())


def mp_testable(test_cls):
    """
    Decorator to copy all the public API for object type test_cls to the decorated class. The decorated
    MPTesterBase subclass will be able to proxy commands to the 'test_cls' instance on a child process/VM via a queue.
    """
    def propogate_cmd(cmd_name):
        def send_cmd(self, *args, **kwargs):
            cmd = (cmd_name, args, kwargs)
            self.socket.send_pyobj(cmd)
        return send_cmd

    def _mp_testable(cls):
        cls.test_cls = test_cls

        # Only copy non-internal and callable methods
        search_scope = [name for name in dir(test_cls) if
                        ((callable(getattr(test_cls, name))) and (len(name) < 2 or name[:2] != '__'))]
        for func in search_scope:
            setattr(cls, func, propogate_cmd(func))

        return cls

    return _mp_testable


def _gen_url(name=''):
    """
    Helper method to generate a random URL for use in a PAIR socket
    """
    rand_num = random.randint(0, pow(2, 16))
    return "ipc://mptest-ipc-{}-{}".format(name, rand_num)


class MPTesterProcess:
    """
    We create the blocking tester object, and configure it with mock objects using a function passed in
    Then, we ensure futures __recv_cmd() to read from cmd_self.socket for proxy'd commands, and __check_assertions() to
    check assertions on a scheduled basis until they complete or until we get a SIG_ABORT from main thread
    """

    def __init__(self, name, url, build_fn, config_fn, assert_fn):
        self.url = url
        self.name = name
        self.config_fn = config_fn
        self.assert_fn = assert_fn
        self.gathered_tasks = None
        self.log = get_logger("TesterProc-{}".format(name))

        # Speculation: New processes has to start with a _new event loop
        self._new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._new_loop)

        self.tester_obj, self.loop, self.tasks = self._build_components(build_fn)

        # Connect to parent process over ipc PAIR self.socket
        self.ctx = zmq.asyncio.Context()
        self.socket = self.ctx.socket(socket_type=zmq.PAIR)
        self.log.info("TestProcess binding to URL {}".format(self.url))
        self.socket.bind(self.url)

        if self.config_fn:
            self.tester_obj = self.config_fn(self.tester_obj)

    def start_test(self):
        """
        Sends ready signal to parent process, and then starts the event self.loop in this process
        """
        assert self.gathered_tasks is None, "start_test can only be called once"
        self.gathered_tasks = asyncio.gather(self._recv_cmd(), *self.tasks)

        self.log.debug("sending ready sig to parent")
        self.socket.send(SIG_RDY)

        try:
            self.log.debug("starting tester proc event loop")
            self.loop.run_until_complete(self.gathered_tasks)
        except Exception as e:
            # If the tasks were canceled internally, then do not run teardown() again
            if type(e) is asyncio.CancelledError:
                self.log.warning("Task(s) cancel detected. Closing event loop.")
                self.loop.stop()
                return

            self.log.error("\n\nException in main TesterProc loop: {}\n\n".format(traceback.format_exc()))
            # self.socket.send_pyobj(SIG_FAIL)
            # self._teardown()

    def _build_components(self, build_fn) -> tuple:
        objs = build_fn()

        # Validate tuple
        assert type(objs) is tuple, "Expected a tuple of length 3 with (tester_obj, loop, tasks)"
        assert len(objs) == 3, "Expected a tuple of length 3 with (tester_obj, loop, tasks)"

        tester_obj, loop, tasks = objs

        # Validate loop
        assert isinstance(loop, asyncio.AbstractEventLoop), \
            "Got {} that isn't an instance of asyncio.AbstractEventLoop".format(loop)
        assert self._new_loop == loop, "Builder object returned _new loop {} that does not match MPTesterProcess's event " \
                                       "loop {}".format(loop, self._new_loop)
        asyncio.set_event_loop(loop)

        # Validate tasks
        assert type(tasks) is list or type(tasks) is tuple, \
            "3rd return val of build_obj must be a list/tuple of tasks... got {} instead".format(tasks)
        # assert len(tasks) >= 1, "Expected at least one task"

        return tester_obj, loop, tasks

    async def _recv_cmd(self):
        """
        Receive commands from the main process and execute them on the tester object. If cmd is equal to ABORT_SIG,
        then we execute __teardown() to stop this self.loop. If its an SIG_START, we start polling for assertions.
        Otherwise, cmd is assumed to be a a tuple of command info of the format
        (func_name: str, args: list, kwargs: dict).
        """
        while True:
            cmd = await self.socket.recv_pyobj()  # recv commands/events from test orchestrator

            # If we got a SIG_ABORT, tear this bish down
            if cmd == SIG_ABORT:
                # self.log.critical("\n!!!!!\nGOT ABORT SIG\n!!!!!\n")
                errs = self._assertions()
                if errs:
                    self.log.critical("\n\n{0}\nASSERTIONS FAILED FOR {2}:\n{1}\n{0}\n".format('!' * 120, errs, self.name))
                self._teardown()
                return

            # If we got a SIG_START, start polling for assertions (if self.assert_fn passed in)
            elif cmd == SIG_START:
                self.log.debug("Got SIG_START from test orchestrator")
                if self.assert_fn:
                    self.log.debug("\nStarting to check assertions every {} seconds\n".format(ASSERTS_POLL_FREQ))
                    asyncio.ensure_future(self._check_assertions())
                continue

            # If msg is not a signal, we assume its a command tuple of form (func, args, kwargs)
            assert len(
                cmd) == 3, "Expected command tuple of len 3 with form (func: str, args: list, kwargs: dict) but " \
                           "got {}".format(cmd)
            func, args, kwargs = cmd

            # Execute cmd in a try/catch, and send a SIG_FAIL to test orchestrator proc if something blow up
            try:
                output = getattr(self.tester_obj, func)(*args, **kwargs)

                # If result is coroutine, run it in the event self.loop
                if output and inspect.iscoroutine(output):
                    result = await asyncio.ensure_future(output)
            # If something blows up, teardown and send a FAIL_SIG to orchestration process
            except Exception as e:
                self.log.error("\n\n TESTER GOT EXCEPTION FROM EXECUTING CMD {}: {}\n\n".format(cmd, traceback.format_exc()))
                self.socket.send(SIG_FAIL)
                self._teardown()
                return

    async def _check_assertions(self):
        """
        Schedule assertion check every TEST_CHECK_FREQ seconds, until either:
            1) The assertions exceed, in which case we send a success signal SUCC_SIG to main thread
            2) The assertions timeout, in which case we send a fail signal FAIL_SIG to main thread
            3) We get can abort cmd (read in __recv_cmd), in which case we send an ABORT_SIG to main thread

        Once one one of these conditions is met, the corresponding signal is sent to the main thread as this
        process calls __teardown() cleans up the event self.loop.
        """
        self.log.debug("Starting assertion checks")

        # Run assertions for until either case (1) or (2) described above occurs
        while True:
            if self._assertions() is None:
                break

            # Sleep until next assertion check
            await asyncio.sleep(ASSERTS_POLL_FREQ)

        # Once out of the assertion checking self.loop, send success to main thread
        self.log.debug("assertions passed! putting ready sig in queue")
        self.socket.send(SIG_SUCC)

    def _teardown(self):
        """
        Stop all tasks and close this processes event self.loop. Invoked after we successfully pass all assertions, or
        timeout.
        """
        self.log.info("Tearing down TesterProcess bound to URL {}".format(self.url))

        self.log.debug("Closing pair socket to MPTest process")
        self.socket.close()

        self.log.debug("Stopping MPTester object tasks")
        self.gathered_tasks.cancel()

    def _assertions(self):
        """
        Helper method to run tester object's assertions, and return the error raised as a string, or None if no
        assertions are raised
        """
        if not self.assert_fn:
            return None

        try:
            self.assert_fn(self.tester_obj)
            return None
        except Exception as e:
            return str(e)


class MPTesterBase:
    """
    TODO docstring
    """
    tester_cls = 'UNSET'

    def __init__(self, config_fn=None, assert_fn=None,  name='TestableProcess', block_until_rdy=True,
                 always_run_as_subproc=False, *args, **kwargs):
        super().__init__()
        self.log = get_logger(name)
        self.name = name

        self.config_fn = config_fn  # Function to configure object with mocks
        self.assert_fn = assert_fn  # Function to run assertions on said mocks

        self.test_proc = None
        self.container_name = None  # Name of the docker container this object is proxying to (if run on VM)

        if 'sk' in kwargs:
            self.sk = kwargs['sk']
            self.vk = wallet.get_vk(kwargs['sk'])

        # Create a wrapper around the build_obj with args and kwargs. We do this b/c this function will actually be
        # invoked in a separate process/machine, thus we need to capture the function call to serialize it and send
        # it across a socket
        build_fn = wrap_fn(type(self).build_obj, *args, name=name, **kwargs)

        self._config_url_and_test_proc(build_fn, always_run_as_subproc)

        # 'socket' is used to proxy commands to blocking object running in a child process (or possibly on a VM)
        self.ctx = zmq.Context()
        self.socket = self.ctx.socket(socket_type=zmq.PAIR)
        self.log.debug("Test Orchestrator connecting to url {}".format(self.url))
        self.socket.connect(self.url)

        # Block this process until we get a ready signal from the subprocess/VM
        if block_until_rdy:
            self.wait_for_test_object()

    def _config_url_and_test_proc(self, build_fn, always_run_as_subproc):
        # Add this object to the registry of testers
        from .mp_test_case import MPTestCase
        MPTestCase.testers.append(self)

        # Create Tester object in a VM
        if MPTestCase.vmnet_test_active and not always_run_as_subproc:
            assert hasattr(MPTestCase, 'ports'), "VMNet test is active, but MPTestCase has no attribute ports"
            assert MPTestCase.ports, "VMNet test is active, but MPTestCase.ports is not set"

            name, ip = MPTestCase.next_container()
            assert name in MPTestCase.ports, "Node named {} not found in ports {}".format(name, MPTestCase.ports)
            assert MPTEST_PORT in MPTestCase.ports[name], "MPTEST_PORT {} not found in docker node {}'s ports {}"\
                                                          .format(MPTEST_PORT, name, MPTestCase.ports[name])

            url = MPTestCase.ports[name][MPTEST_PORT]  # URL the orchestration node should connect to
            url = url.replace('localhost', '127.0.0.1') # Adjust localhost to 127.0.0.1
            url = "tcp://{}".format(url)

            remote_url = "tcp://{}:{}".format(ip, MPTEST_PORT)  # URL the remote node should bind to
            self.container_name = name
            self.url = url
            self.ip = ip

            self.log.notice("Creating node named {} in a docker container with name {} and ip {}".format(self.name, name, self.ip))

            env_vars = {}
            if os.getenv('CIRLCECI'):
                env_vars['CIRCLECI'] = '1'

            runner_func = wrap_fn(_run_test_proc, self.name, remote_url, build_fn, self.config_fn,
                                  self.assert_fn, env_vars=env_vars, log_lvl=MPTestCase.log_lvl)

            # TODO -- will i need a ton of imports and stuff to make this run smoothly...?
            MPTestCase.execute_python(name, runner_func, async=True)

        # Create Tester object in a Subprocess
        else:
            self.url = _gen_url(self.name)
            self.log.notice("Creating node named {} in a subprocess with IPC url {}".format(self.name, self.url))

            self.test_proc = LProcess(target=_run_test_proc, name=self.name, args=(self.name, self.url, build_fn,
                                                                                   self.config_fn, self.assert_fn,
                                                                                   ))
            self.test_proc.start()

    def reconnect(self):
        old_url = self.url
        self.log.notice("Host machine disconnecting to old URL {}".format(old_url))
        self.socket.disconnect(old_url)

        url = BaseNetworkTestCase.ports[self.container_name][MPTEST_PORT]  # URL the orchestration node should connect to
        url = url.replace('localhost', '127.0.0.1')  # Adjust localhost to 127.0.0.1
        url = "tcp://{}".format(url)
        self.log.debug("Host machine changing old URL {} to _new URL {} for container {}".format(old_url, url, self.container_name))
        self.url = url

        self.log.notice("Host machine reconnecting to URL {}".format(self.url))
        self.socket.connect(self.url)

    def wait_for_test_object(self):
        self.log.info("Tester waiting for rdy sig from test process...")
        msg = self.socket.recv()
        assert msg == SIG_RDY, "Got msg from child thread {} but expected SIG_RDY".format(msg)
        self.log.info("GOT RDY SIG: {}".format(msg))

    @classmethod
    def build_obj(cls, *args, **kwargs) -> tuple:
        """
        Override to define how the blocking object should be initialized. Must return 2 element tuple, in the order
        <loop> (an EventLoop instance), and <object> (an instance of the blocking object to test).
        :return: Tuple of the form (loop, object_instance)

        It is assumed the object being returning is passed an event loop into its constructor, which it uses internally
        to schedule all its events.
        """
        raise NotImplementedError

    def teardown(self):
        self.log.debug("{} tearing down...".format(self.name))

        self.socket.close()

        if self.test_proc:
            self.log.debug("Joining tester terminate {}...".format(self.name))
            self.test_proc.terminate()
            self.test_proc.join()
            self.log.debug("Tester Proc {} joined".format(self.name))

        if self.container_name:
            # TODO clean up container???
            pass

        self.log.debug("{} done tearing down.".format(self.name))

    def __repr__(self):
        return '<' + self.name + "  " + str(type(self)) + '>'



import os, sys

# Add /messages/capnp to Python path. We need these loaded for capnp magic imports
sys.path.append(os.path.dirname(__file__) + '/messages/capnp')

# Set the decimal precision for floating point arithmetic
# from decimal import getcontext
# getcontext().prec = DECIMAL_PRECISION



# Number of seconds Masternode will wait in respective states before he times out
STAGING_TIMEOUT = 30
NEW_BLOCK_TIMEOUT = 8
FETCH_BLOCK_TIMEOUT = 24

# Quorums
TOTAL_MN = 12
QUORUM = 3
MN_ID = 1

# Web Server
NUM_WORKERS = 2  # Number of Sanic worker procs
NONCE_EXPIR = 90  # Number of seconds before a nonce requested by a user expires

# Storage
REP_FACTOR = 3
MAX_BLOCK = 100
MN_BLK_DATABASE = 'mn_store'
MN_INDEX_DATABASE = 'mn_index'
TEST_HOOK = False

# Block Aggregator
BLOCK_TIMEOUT_POLL = 2.0  # how often BlockAggregator should check if the current block being produced has timed out



from cilantro_ee.core.sockets.socket_book import SocketBook
from cilantro_ee.services.storage.vkbook import PhoneBook
from cilantro_ee.core.top import TopBlockManager
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.messages.message import Message, MessageType
from cilantro_ee.core.canonical import verify_block
from cilantro_ee.core.sockets.services import get, defer
from cilantro_ee.services.storage.master import CilantroStorageDriver
import zmq.asyncio
import asyncio
from collections import Counter
import time


class ConfirmationCounter(Counter):
    def top_item(self):
        return self.most_common()[0][0]

    def top_count(self):
        if len(self.most_common()) == 0:
            return 0
        return self.most_common()[0][1]


# Open a socket and to listen for new block notifications
class BlockFetcher:
    def __init__(self, wallet: Wallet,
                 ctx: zmq.Context,
                 blocks: CilantroStorageDriver=None,
                 top=TopBlockManager(),
                 state=MetaDataStorage(),
                 masternode_sockets=SocketBook(None, PhoneBook.contract.get_masternodes)):

        self.masternodes = masternode_sockets
        self.top = top
        self.wallet = wallet
        self.ctx = ctx
        self.blocks = blocks
        self.state = state

    # Change to max received
    async def find_missing_block_indexes(self, confirmations=3, timeout=3000):
        await self.masternodes.refresh()

        responses = ConfirmationCounter()

        futures = []
        # Fire off requests to masternodes on the network
        for master in self.masternodes.sockets.values():
            f = asyncio.ensure_future(self.get_latest_block_height(master))
            futures.append(f)

        # Iterate through the status of the
        now = time.time()
        while responses.top_count() < confirmations or time.time() - now > timeout:
            await defer()
            for f in futures:
                if f.done():
                    responses.update([f.result()])

                    # Remove future
                    futures.remove(f)

        return responses.top_item()

    async def get_latest_block_height(self, socket):
        request = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                      msg_type=MessageType.LATEST_BLOCK_HEIGHT_REQUEST,
                                                      timestamp=int(time.time()))

        response = await get(socket_id=socket, msg=request, ctx=self.ctx, timeout=3000, retries=0, dealer=True)

        if response is not None:
            _, unpacked, _, _, _ = Message.unpack_message_2(response)

            return unpacked.blockHeight

    async def get_block_from_master(self, i: int, socket):
        request = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                      msg_type=MessageType.BLOCK_DATA_REQUEST,
                                                      blockNum=i)

        response = await get(socket_id=socket, msg=request, ctx=self.ctx, timeout=3000, retries=0, dealer=True)

        if response is not None:
            msg_type, unpacked, _, _, _ = Message.unpack_message_2(response)

            if msg_type == MessageType.BLOCK_DATA:
                return unpacked

    async def find_valid_block(self, i, latest_hash, timeout=3000):
        await self.masternodes.refresh()

        block_found = False
        block = None

        futures = []
        # Fire off requests to masternodes on the network
        for master in self.masternodes.sockets.values():
            f = asyncio.ensure_future(self.get_block_from_master(i, master))
            futures.append(f)

        # Iterate through the status of the
        now = time.time()
        while not block_found or time.time() - now > timeout:
            await defer()
            for f in futures:
                if f.done():
                    block = f.result()
                    block_found = verify_block(subblocks=block.subBlocks,
                                               previous_hash=latest_hash,
                                               proposed_hash=block.blockHash)
                    futures.remove(f)

        return block

    async def fetch_blocks(self, latest_block_available=0):
        latest_block_stored = self.top.get_latest_block_number()
        latest_hash = self.top.get_latest_block_hash()

        if latest_block_available <= latest_block_stored:
            return

        for i in range(latest_block_stored, latest_block_available + 1):
            block = await self.find_valid_block(i, latest_hash)

            if block is not None:
                block_dict = {
                    'blockHash': block.blockHash,
                    'blockNum': i,
                    'blockOwners': [m for m in block.blockOwners],
                    'prevBlockHash': latest_hash,
                    'subBlocks': [s for s in block.subBlocks]
                }

                # Only store if master, update state if master or delegate

                if self.blocks is not None:
                    self.blocks.put(block_dict)

                self.state.update_with_block(block_dict)
                self.top.set_latest_block_hash(block.blockHash)
                self.top.set_latest_block_number(i)

                latest_hash = self.top.get_latest_block_hash()
            else:
                raise Exception('Could not find block with index {}. Catchup failed.'.format(i))

    async def sync(self):
        current_height = await self.find_missing_block_indexes()
        latest_block_stored = self.top.get_latest_block_number()

        while current_height < latest_block_stored:
            await self.fetch_blocks(current_height)
            current_height = await self.find_missing_block_indexes()

    async def sync_blocks_with_state(self):
        if self.blocks is None:
            return

        last_block = self.blocks.get_last_n(1, CilantroStorageDriver.INDEX)[0]
        last_stored_block_num = last_block.get('blockNum')
        last_state_block_num = self.top.get_latest_block_number()

        while last_state_block_num < last_stored_block_num:
            last_state_block_num += 1
            block_dict = self.blocks.get_block(last_state_block_num)

            self.state.update_with_block(block_dict)

# struct BlockData {
#     blockHash @0 :Data;
#     blockNum @1 :UInt32;
#     blockOwners @2 :List(Data);
#     prevBlockHash @3 :Data;
#     subBlocks @4 :List(SB.SubBlock);
# }



from cilantro_ee.core.utils.transaction import TransactionBuilder, verify_packed_tx
from unittest import TestCase
from cilantro_ee.core.crypto.wallet import Wallet
import os
from cilantro_ee.messages import capnp as schemas

import capnp
transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')


class TestContractTransaction(TestCase):
    def test_init(self):
        TransactionBuilder('blah', 123, 'blah', 'blah', 'blah', {'something': 123})

    def test_signing_flips_true(self):
        w = Wallet()
        tx = TransactionBuilder(w.verifying_key().hex(),
                                1000000, 'currency', 'transfer', 'test',
                                {'amount': 123})

        self.assertFalse(tx.tx_signed)

        tx.sign(w.signing_key().hex())

        self.assertTrue(tx.tx_signed)

    def test_generate_proof_flips_true(self):
        w = Wallet()
        tx = TransactionBuilder(w.verifying_key().hex(),
                                1000000, 'currency', 'transfer', 'test',
                                {'amount': 123})

        self.assertFalse(tx.proof_generated)

        tx.generate_proof()

        self.assertTrue(tx.proof_generated)

    def test_serialize_if_not_signed_returns_none(self):
        w = Wallet()
        tx = TransactionBuilder(w.verifying_key().hex(),
                                1000000, 'currency', 'transfer', 'test',
                                {'amount': 123})

        self.assertIsNone(tx.serialize())

    def test_serialize_returns_bytes(self):
        w = Wallet()
        tx = TransactionBuilder(w.verifying_key().hex(),
                                1000000, 'currency', 'transfer', 'test',
                                {'amount': 123})

        tx.sign(w.signing_key().hex())

        tx_packed = tx.serialize()

        self.assertTrue(verify_packed_tx(w.verifying_key(), tx_packed))

    def test_bad_bytes_returns_false_on_verify(self):
        w = Wallet()
        b = b'bad'
        self.assertFalse(verify_packed_tx(w.verifying_key(), b))

    def test_passing_float_in_contract_kwargs_raises_assertion(self):
        w = Wallet()
        with self.assertRaises(AssertionError):
            TransactionBuilder(w.verifying_key().hex(),
                               1000000, 'currency', 'transfer', 'test',
                               {'amount': 123.00})

    def test_passing_non_supported_type_in_contract_kwargs_raises_assertion(self):
        w = Wallet()
        with self.assertRaises(AssertionError):
            TransactionBuilder(w.verifying_key().hex(),
                               1000000, 'currency', 'transfer', 'test',
                               {'amount': ['b']})



from cilantro_ee.core.logger import get_logger
import json, math, os
from cilantro_ee.utils.test.testnet_config import get_testnet_json_path

"""
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
TO THE POOR SOUL WHO IS LOOKING AT THIS FILE:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

I'm sorry. I really am. I am tempted to shift blame here, but really I must accept responsibility for the utter
catastrophe that you about to witness below. Buckle up for this bumpy ride down jank code lane. 

If you are reading this code thinking "tf is he doing?" --the answer
is it's all a massive hack. This entire "constitution seeding" business needs to be completely refactored. 
Only tricky bit is the solution must allow for tests (unit tests) to use different constitution files for difference 
situations. This is probably best achieved using Mock objects, and maybe some kind of helper to orchestrate the VK
mocking.

Proposal for proper way to do it:
Fuck off with all these global variables, and just use the class variables on VKBook. Have a helper function to
seed VKBook with different data, which sets the class variables on VKBook to whatever you need for your testing
environment.

Best wishes, 
Davis
"""

TESTNET_JSON_PATH = get_testnet_json_path()
log = get_logger("TestnetBuilder")


# DEBUG -- TODO DELETE
# print("\n\n\n\n\n TESTNET JSON PATH: {} \n\n\n\n\n".format(TESTNET_JSON_PATH))
# END DEBUG


_MASTERNODES = None
_WITNESSES = None
_DELEGATES = None

TESTNET_MASTERNODES = []
TESTNET_WITNESSES = []
TESTNET_DELEGATES = []

r = None  # Replication Factor


def set_testnet_nodes():
    global _MASTERNODES, _DELEGATES, _WITNESSES, TESTNET_MASTERNODES, TESTNET_WITNESSES, TESTNET_DELEGATES, \
        MN_WITNESS_MAP, WITNESS_MN_MAP, r

    with open(TESTNET_JSON_PATH, 'r') as f:
        _dat_good_json = json.load(f)

    _MASTERNODES = _dat_good_json['masternodes']
    _WITNESSES = _dat_good_json['witnesses']
    _DELEGATES = _dat_good_json['delegates']
    assert len(_WITNESSES) % len(
        _MASTERNODES) == 0, "# of witnesses must be divisible by # of masternodes, but got {} ma" \
                            "sternodes and {} witnesses".format(len(_MASTERNODES), len(_WITNESSES))

    TESTNET_MASTERNODES = [{'sk': node['sk'], 'vk': node['vk']} for node in _MASTERNODES]
    TESTNET_WITNESSES = [{'sk': node['sk'], 'vk': node['vk']} for node in _WITNESSES]
    TESTNET_DELEGATES = [{'sk': node['sk'], 'vk': node['vk']} for node in _DELEGATES]

    r = len(_WITNESSES) // len(_MASTERNODES)  # replication factor
    MN_WITNESS_MAP = {}  # Map of masternodes --> responsible witness set
    WITNESS_MN_MAP = {}  # inverse of map above

    # Build MN_WITNESS_MAP/WITNESS_MN_MAP
    for i, mn in enumerate(_MASTERNODES):
        mn_vk = mn['vk']
        witnesses = [node['vk'] for node in _WITNESSES[i * r:i * r + r]]

        MN_WITNESS_MAP[mn_vk] = witnesses
        for w in witnesses:
            WITNESS_MN_MAP[w] = mn_vk


def get_testnet_nodes():
    global _MASTERNODES, _DELEGATES, _WITNESSES, TESTNET_MASTERNODES, TESTNET_WITNESSES, TESTNET_DELEGATES, \
        MN_WITNESS_MAP, WITNESS_MN_MAP, r

    with open(TESTNET_JSON_PATH, 'r') as f:
        _dat_good_json = json.load(f)

    _MASTERNODES = _dat_good_json['masternodes']
    _WITNESSES = _dat_good_json['witnesses']
    _DELEGATES = _dat_good_json['delegates']
    assert len(_WITNESSES) % len(
        _MASTERNODES) == 0, "# of witnesses must be divisible by # of masternodes, but got {} ma" \
                            "sternodes and {} witnesses".format(len(_MASTERNODES), len(_WITNESSES))

    TESTNET_MASTERNODES = [{'sk': node['sk'], 'vk': node['vk']} for node in _MASTERNODES]
    TESTNET_WITNESSES = [{'sk': node['sk'], 'vk': node['vk']} for node in _WITNESSES]
    TESTNET_DELEGATES = [{'sk': node['sk'], 'vk': node['vk']} for node in _DELEGATES]

    r = len(_WITNESSES) // len(_MASTERNODES)  # replication factor
    MN_WITNESS_MAP = {}  # Map of masternodes --> responsible witness set
    WITNESS_MN_MAP = {}  # inverse of map above

    # Build MN_WITNESS_MAP/WITNESS_MN_MAP
    for i, mn in enumerate(_MASTERNODES):
        mn_vk = mn['vk']
        witnesses = [node['vk'] for node in _WITNESSES[i * r:i * r + r]]

        MN_WITNESS_MAP[mn_vk] = witnesses
        for w in witnesses:
            WITNESS_MN_MAP[w] = mn_vk

    return TESTNET_MASTERNODES, TESTNET_WITNESSES, TESTNET_DELEGATES



set_testnet_nodes()



from unittest import TestCase
from cilantro_ee.services.overlay.discovery import *
import zmq
import zmq.asyncio
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.sockets.services import _socket
from time import sleep

TIME_UNIT = 0.01


def run_silent_loop(tasks, s=TIME_UNIT):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(tasks)
    except RuntimeError as e:
        pass
    sleep(s)


async def timeout_bomb(s=TIME_UNIT*2):
    await asyncio.sleep(s)
    asyncio.get_event_loop().close()


class TestDiscoveryServer(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_init(self):
        DiscoveryServer(_socket('tcp://127.0.0.1:10999'), Wallet(), b'blah')

    def test_run_server(self):
        d = DiscoveryServer(_socket('tcp://127.0.0.1:10999'), Wallet(), b'blah')

        tasks = asyncio.gather(timeout_bomb(), d.serve())
        run_silent_loop(tasks)

        d.ctx.destroy()

    def test_send_message_to_discovery(self):
        address = 'tcp://127.0.0.1:10999'

        d = DiscoveryServer(_socket('tcp://127.0.0.1:10999'), Wallet(), b'blah', ctx=self.ctx)

        async def ping(msg, sleep):
            await asyncio.sleep(sleep)
            socket = self.ctx.socket(zmq.REQ)
            socket.setsockopt(zmq.LINGER, 20)
            socket.connect(address)
            socket.send(msg)

            result = await socket.recv()
            self.assertTrue(result)

        tasks = asyncio.gather(ping(b'', TIME_UNIT), d.serve(), timeout_bomb())
        run_silent_loop(tasks)

    def test_verify_vk_pepper_correct_vk_pepper_message(self):
        wallet = Wallet()
        vk = wallet.verifying_key()
        pepper = b'TESTING_PEPPER'

        pepper_msg = vk + wallet.sign(pepper)

        self.assertTrue(verify_vk_pepper(pepper_msg, pepper))

    def test_verify_vk_pepper_wrong_vk_pepper_message(self):
        wallet = Wallet()
        vk = wallet.verifying_key()
        pepper = b'TESTING_PEPPER'

        pepper_msg = vk + wallet.sign(pepper)

        self.assertFalse(verify_vk_pepper(pepper_msg, b'WRONG_PEPPER'))

    def test_verify_vk_pepper_length_greater_than_32_bytes_assertion(self):
        with self.assertRaises(AssertionError):
            verify_vk_pepper(b'0', b'WRONG_PEPPER')

    def test_discovery_server_returns_correct_vk_and_pepper(self):
        address = 'tcp://127.0.0.1:10999'

        wallet = Wallet()

        d = DiscoveryServer(_socket('tcp://127.0.0.1:10999'), wallet, b'CORRECT_PEPPER', ctx=self.ctx)

        async def ping(msg, sleep):
            await asyncio.sleep(sleep)
            socket = self.ctx.socket(zmq.REQ)
            socket.setsockopt(zmq.LINGER, 20)
            socket.connect(address)
            socket.send(msg)

            result = await socket.recv()
            self.assertTrue(verify_vk_pepper(result, b'CORRECT_PEPPER'))

        tasks = asyncio.gather(ping(b'', TIME_UNIT), d.serve(), timeout_bomb())
        run_silent_loop(tasks)

    def test_discovery_server_returns_wrong_vk_and_pepper(self):

        address = 'inproc://testing'

        wallet = Wallet()

        d = DiscoveryServer(_socket('tcp://127.0.0.1:10999'), wallet, b'WRONG_PEPPER', ctx=self.ctx)

        async def ping(msg, sleep):
            await asyncio.sleep(sleep)
            socket = self.ctx.socket(zmq.REQ)
            socket.setsockopt(zmq.LINGER, 20)
            socket.connect(address)
            socket.send(msg)

            result = await socket.recv()
            self.assertFalse(verify_vk_pepper(result, b'CORRECT_PEPPER'))

        tasks = asyncio.gather(ping(b'', TIME_UNIT), d.serve(), timeout_bomb())
        run_silent_loop(tasks)

    def test_discover_server_gets_correct_vk_from_msg_decoding(self):
        address = 'tcp://127.0.0.1:10999'

        wallet = Wallet()

        d = DiscoveryServer(_socket('tcp://127.0.0.1:10999'), wallet, b'CORRECT_PEPPER', ctx=self.ctx)

        async def ping(msg, sleep):
            await asyncio.sleep(sleep)
            socket = self.ctx.socket(zmq.REQ)
            socket.setsockopt(zmq.LINGER, 20)
            socket.connect(address)
            socket.send(msg)

            result = await socket.recv()

            vk, _ = unpack_pepper_msg(result)

            self.assertEqual(vk, wallet.verifying_key())

        tasks = asyncio.gather(ping(b'', TIME_UNIT), d.serve(), timeout_bomb())
        run_silent_loop(tasks)

    def test_async_ping_timeout_occurs_if_ip_isnt_online(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(ping(_socket('inproc://dontexist'),
                                     pepper=b'DOESNT_MATTER',
                                     ctx=self.ctx,
                                     timeout=50,)
                                )

    def test_one_vk_returned_if_one_ip_is_online(self):
        wallet = Wallet()

        d = DiscoveryServer(_socket('tcp://127.0.0.1:10999'), wallet, b'CORRECT_PEPPER', ctx=self.ctx)

        success_task = ping(_socket('tcp://127.0.0.1:10999'),
                            pepper=b'CORRECT_PEPPER',
                            ctx=self.ctx,
                            timeout=300)

        failure_task = ping(_socket('tcp://127.0.0.1:20999'),
                            pepper=b'CORRECT_PEPPER',
                            ctx=self.ctx,
                            timeout=300)

        async def stop_server(timeout):
            await asyncio.sleep(timeout)
            d.stop()

        tasks = asyncio.gather(success_task, failure_task, d.serve(), stop_server(0.3))

        loop = asyncio.get_event_loop()
        results = loop.run_until_complete(tasks)

        vk_ip1, vk_ip2, _, _ = results

        _, vk1 = vk_ip1
        _, vk2 = vk_ip2

        self.assertEqual(vk1.hex(), wallet.verifying_key().hex())
        self.assertIsNone(vk2)

    def test_discover_nodes_found_one(self):
        address = _socket('tcp://127.0.0.1:10999')

        wallet = Wallet()

        d = DiscoveryServer(address, wallet, b'CORRECT_PEPPER', ctx=self.ctx)

        async def stop_server(timeout):
            await asyncio.sleep(timeout)
            d.stop()

        tasks = asyncio.gather(
            discover_nodes(ip_list=[address], pepper=b'CORRECT_PEPPER', ctx=self.ctx),
            d.serve(),
            stop_server(0.2)
        )

        loop = asyncio.get_event_loop()
        results = loop.run_until_complete(tasks)

        r = results[0]

        self.assertEqual(r[str(address)], wallet.verifying_key().hex())

    def test_discover_nodes_found_three(self):
        addresses = [_socket('tcp://127.0.0.1:10999'), _socket('tcp://127.0.0.1:11999'), _socket('tcp://127.0.0.1:12999')]
        wallets = [Wallet(), Wallet(), Wallet()]
        pepper = b'CORRECT_PEPPER'
        server_timeout = 0.3

        servers = [DiscoveryServer(addresses[0], wallets[0], pepper, ctx=self.ctx),
                   DiscoveryServer(addresses[1], wallets[1], pepper, ctx=self.ctx),
                   DiscoveryServer(addresses[2], wallets[2], pepper, ctx=self.ctx)]

        async def stop_server(s, timeout):
            await asyncio.sleep(timeout)
            s.stop()

        tasks = asyncio.gather(
            servers[0].serve(),
            servers[1].serve(),
            servers[2].serve(),
            stop_server(servers[0], server_timeout),
            stop_server(servers[1], server_timeout),
            stop_server(servers[2], server_timeout),
            discover_nodes(ip_list=addresses, pepper=pepper, ctx=self.ctx)
        )

        loop = asyncio.get_event_loop()
        results = loop.run_until_complete(tasks)

        r = results[-1]

        self.assertEqual(r[str(addresses[0])], wallets[0].verifying_key().hex())
        self.assertEqual(r[str(addresses[1])], wallets[1].verifying_key().hex())
        self.assertEqual(r[str(addresses[2])], wallets[2].verifying_key().hex())

    def test_discover_nodes_found_two_out_of_three(self):
        addresses = [_socket('tcp://127.0.0.1:10999'), _socket('tcp://127.0.0.1:11999'), _socket('tcp://127.0.0.1:12999')]
        addresses_wrong = [_socket('tcp://127.0.0.1:10999'), _socket('tcp://127.0.0.1:11999'), _socket('tcp://127.0.0.1:13999')]
        wallets = [Wallet(), Wallet(), Wallet()]
        pepper = b'CORRECT_PEPPER'
        server_timeout = 1

        servers = [DiscoveryServer(addresses[0], wallets[0], pepper, ctx=self.ctx),
                   DiscoveryServer(addresses[1], wallets[1], pepper, ctx=self.ctx),
                   DiscoveryServer(addresses[2], wallets[2], pepper, ctx=self.ctx)]

        async def stop_server(s, timeout):
            await asyncio.sleep(timeout)
            s.stop()

        tasks = asyncio.gather(
            servers[0].serve(),
            servers[1].serve(),
            servers[2].serve(),
            stop_server(servers[0], server_timeout),
            stop_server(servers[1], server_timeout),
            stop_server(servers[2], server_timeout),
            discover_nodes(ip_list=addresses_wrong, pepper=pepper, ctx=self.ctx)
        )

        loop = asyncio.get_event_loop()
        results = loop.run_until_complete(tasks)

        r = results[-1]

        print(r)

        self.assertEqual(r[str(addresses[0])], wallets[0].verifying_key().hex())
        self.assertEqual(r[str(addresses[1])], wallets[1].verifying_key().hex())
        self.assertIsNone(r.get(str(addresses[2])))

    def test_discover_nodes_none_found(self):
        addresses = [_socket('tcp://127.0.0.1:10999'), _socket('tcp://127.0.0.1:11999'), _socket('tcp://127.0.0.1:12999')]
        addresses_wrong = [_socket('tcp://127.0.0.1:15999'), _socket('tcp://127.0.0.1:14999'), _socket('tcp://127.0.0.1:13999')]
        wallets = [Wallet(), Wallet(), Wallet()]
        pepper = b'CORRECT_PEPPER'
        server_timeout = 1

        servers = [DiscoveryServer(addresses[0], wallets[0], pepper, ctx=self.ctx),
                   DiscoveryServer(addresses[1], wallets[1], pepper, ctx=self.ctx),
                   DiscoveryServer(addresses[2], wallets[2], pepper, ctx=self.ctx)]

        async def stop_server(s, timeout):
            await asyncio.sleep(timeout)
            s.stop()

        tasks = asyncio.gather(
            servers[0].serve(),
            servers[1].serve(),
            servers[2].serve(),
            stop_server(servers[0], server_timeout),
            stop_server(servers[1], server_timeout),
            stop_server(servers[2], server_timeout),
            discover_nodes(ip_list=addresses_wrong, pepper=pepper, ctx=self.ctx, timeout=500, retries=3)
        )

        loop = asyncio.get_event_loop()
        results = loop.run_until_complete(tasks)

        r = results[-1]

        self.assertIsNone(r.get(addresses[0]))
        self.assertIsNone(r.get(addresses[1]))
        self.assertIsNone(r.get(addresses[2]))



from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('2-2-0.json')

from vmnet.testcase import BaseNetworkTestCase
import unittest, cilantro_ee
from os.path import join, dirname
from cilantro_ee.utils.test.mp_test_case import vmnet_test
from cilantro_ee.utils.test.god import God
from cilantro_ee.core.logger import get_logger

LOG_LEVEL = 0


def wrap_func(fn, *args, **kwargs):
    def wrapper():
        return fn(*args, **kwargs)
    return wrapper


def run_mn(slot_num):
    from cilantro_ee.utils.factory import NodeFactory
    from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
    import os

    # overwrite_logger_level(logging.WARNING)
    # overwrite_logger_level(21)
    # overwrite_logger_level(11)
    from cilantro_ee.core.logger.base import get_logger
    log = get_logger("MN CREATOR")
    log.important("creating")

    ip = os.getenv('HOST_IP')
    sk = TESTNET_MASTERNODES[slot_num]['sk']
    NodeFactory.run_masternode(ip=ip, signing_key=sk, reset_db=True)


def run_witness(slot_num):
    from cilantro_ee.utils.factory import NodeFactory
    from cilantro_ee.constants.testnet import TESTNET_WITNESSES
    import os

    # overwrite_logger_level(logging.WARNING)
    # overwrite_logger_level(21)
    # overwrite_logger_level(11)

    w_info = TESTNET_WITNESSES[slot_num]
    w_info['ip'] = os.getenv('HOST_IP')

    NodeFactory.run_witness(ip=w_info['ip'], signing_key=w_info['sk'], reset_db=True)


def dump_it(volume, delay=0):
    from cilantro_ee.utils.test.god import God
    from cilantro_ee.core.logger import overwrite_logger_level
    import logging

    overwrite_logger_level(logging.WARNING)
    God.dump_it(volume=volume, delay=delay)


class TestManualDump(BaseNetworkTestCase):

    VOLUME = 1200  # Number of transactions to dump
    config_file = join(dirname(cilantro_ee.__path__[0]), 'vmnet_configs', 'cilantro_ee-2-2-0-bootstrap.json')
    PROFILE_TYPE = None

    @vmnet_test(run_webui=True)
    def test_dump(self):
        log = get_logger("Dumpatron")
        log.important3("DUMPATRON REPORTING FOR DUTY")

        # Bootstrap master
        for i, nodename in enumerate(self.groups['masternode']):
            self.execute_python(nodename, wrap_func(run_mn, i), async=True, profiling=self.PROFILE_TYPE)

        # Bootstrap witnesses
        for i, nodename in enumerate(self.groups['witness']):
            self.execute_python(nodename, wrap_func(run_witness, i), async=True, profiling=self.PROFILE_TYPE)

        while True:
            user_input = input("Enter an integer representing the # of transactions to dump, or 'x' to quit.")
            if user_input.lower() == 'x':
                log.debug("Termination input detected. Breaking")
                break

            vol = int(user_input) if user_input.isdigit() else self.VOLUME
            log.important3("Dumpatron dumping {} transactions!".format(vol))
            self.execute_python('mgmt', wrap_func(dump_it, volume=vol), async=True, profiling=self.PROFILE_TYPE)

        log.important3("Dumpatron initiating system teardown")
        God.teardown_all("http://{}".format(self.ports['masternode']['8080']))


if __name__ == '__main__':
    unittest.main()



from cilantro_ee.constants import conf
from cilantro_ee.constants.ports import DHT_PORT, DISCOVERY_PORT, EVENT_PORT
from cilantro_ee.constants.overlay_network import PEPPER
from cilantro_ee.services.overlay import discovery
from cilantro_ee.core.sockets import services
from cilantro_ee.core.crypto.wallet import Wallet

from copy import deepcopy
from functools import partial
import asyncio
import json
import zmq
from cilantro_ee.core.logger.base import get_logger

import random

log = get_logger('NetworkService')


def bytes_from_ip_string(ip: str):
    b = ip.split('.')
    bb = [int(i) for i in b]
    return bytes(bb)


class KTable:
    def __init__(self, data: dict, initial_peers={}, response_size=10):
        self.data = data
        self.peers = initial_peers
        self.response_size = response_size

    @staticmethod
    def distance(string_a, string_b):
        int_val_a = int(string_a.encode().hex(), 16)
        int_val_b = int(string_b.encode().hex(), 16)
        return int_val_a ^ int_val_b

    def find(self, key):
        if key in self.data:
            return self.data
        elif key in self.peers:
            return {
                key: self.peers[key]
            }
        else:
            # Do an XOR sort on all the keys to find neighbors
            sort_func = partial(self.distance, string_b=key)
            closest_peer_keys = sorted(self.peers.keys(), key=sort_func)

            # Only keep the response size number
            closest_peer_keys = closest_peer_keys[:self.response_size]

            # Dict comprehension
            neighbors = {k: self.peers[k] for k in closest_peer_keys}

            return neighbors


class PeerServer(services.RequestReplyService):
    def __init__(self, socket_id: services.SocketStruct,
                 event_port: int,
                 discovery_port: int,
                 table: KTable, wallet: Wallet, ctx=zmq.Context,
                 linger=2000, poll_timeout=3000, pepper=PEPPER.encode()):

        super().__init__(socket_id=socket_id,
                         wallet=wallet,
                         ctx=ctx,
                         linger=linger,
                         poll_timeout=poll_timeout)

        self.table = table

        self.event_address = 'tcp://*:{}'.format(event_port)
        self.event_service = services.SubscriptionService(ctx=self.ctx)

        self.event_publisher = self.ctx.socket(zmq.PUB)
        self.event_publisher.bind(self.event_address)

        self.event_queue_loop_running = False

        self.discovery_port = discovery_port

    def handle_msg(self, msg):
        msg = msg.decode()
        command, args = json.loads(msg)

        if command == 'find':
            response = self.table.find(args)
            response = json.dumps(response, cls=services.SocketEncoder).encode()
            return response
        if command == 'join':
            vk, ip = args # unpack args
            asyncio.ensure_future(self.handle_join(vk, ip))
            return None
        #if command == 'ping':
        #    return self.ping_response

    async def handle_join(self, vk, ip):
        result = self.table.find(vk)

        if vk not in result or result[vk] != ip:
            # Ping discovery server
            _, responded_vk = await discovery.ping(services._socket(ip),
                                    pepper=PEPPER.encode(), ctx=self.ctx, timeout=1000)

            await asyncio.sleep(0)
            if responded_vk is None:
                return

            if responded_vk.hex() == vk:
                # Valid response
                self.table.peers[vk] = ip

                # Publish a message that a _new node has joined
                msg = ['join', (vk, ip)]
                jmsg = json.dumps(msg, cls=services.SocketEncoder).encode()
                await self.event_publisher.send(jmsg)

                second_msg = json.dumps({'event': 'node_online', 'vk': vk, 'ip': services._socket(ip).id}, cls=services.SocketEncoder).encode()
                await self.event_publisher.send(second_msg)

    async def process_event_subscription_queue(self):
        self.event_queue_loop_running = True

        while self.event_queue_loop_running:
            if len(self.event_service.received) > 0:
                message, sender = self.event_service.received.pop(0)
                msg = json.loads(message.decode())

                # Ignore event dictionaries
                if isinstance(msg, dict):
                    continue

                command, args = msg
                vk, ip = args

                if command == 'join':
                    asyncio.ensure_future(self.handle_join(vk=vk, ip=ip))

                elif command == 'leave':
                    # Ping to make sure the node is actually offline
                    _, responded_vk = await discovery.ping(ip, pepper=PEPPER.encode(),
                                                           ctx=self.ctx, timeout=1000)

                    # If so, remove it from our table
                    if responded_vk is None:
                        del self.table[vk]

            await asyncio.sleep(0)

    async def start(self):
        asyncio.ensure_future(asyncio.gather(
            self.serve(),
            self.event_service.serve(),
            self.process_event_subscription_queue()
        ))
        log.info('Peer services running on {}'.format(self.address))

    def stop(self):
        self.running = False
        self.event_queue_loop_running = False
        self.event_service.running = False


class Network:
    def __init__(self, wallet,
                 peer_service_port: int=DHT_PORT,
                 event_publisher_port: int=EVENT_PORT,
                 discovery_port: int=DISCOVERY_PORT,
                 ctx=zmq.asyncio.Context(),
                 ip=conf.HOST_IP,
                 bootnodes=conf.BOOT_DELEGATE_IP_LIST + conf.BOOT_MASTERNODE_IP_LIST,
                 initial_mn_quorum=1,
                 initial_del_quorum=1,
                 mn_to_find=[],
                 del_to_find=[]):

        # General Instance Variables
        self.wallet = wallet
        self.ctx = ctx

        self.bootnodes = bootnodes
        self.ip = ip

        # Peer Service Constants
        self.peer_service_address = services.SocketStruct(services.Protocols.TCP, '*', peer_service_port)

        data = {
            self.wallet.verifying_key().hex(): ip
        }
        self.table = KTable(data=data)

        self.event_publisher_address = 'tcp://*:{}'.format(event_publisher_port)
        self.peer_service_port = peer_service_port
        self.peer_service = PeerServer(self.peer_service_address,
                                       event_port=event_publisher_port,
                                       discovery_port=discovery_port,
                                       table=self.table, wallet=self.wallet, ctx=self.ctx)

        self.discovery_port = discovery_port
        self.discovery_server_address = services.SocketStruct(services.Protocols.TCP, '*', self.discovery_port)
        self.discovery_server = discovery.DiscoveryServer(self.discovery_server_address,
                                                          wallet=self.wallet,
                                                          pepper=PEPPER.encode(),
                                                          ctx=self.ctx)

        # Quorum Constants
        self.initial_mn_quorum = initial_mn_quorum
        self.initial_del_quorum = initial_del_quorum
        self.mn_to_find = mn_to_find
        self.del_to_find = del_to_find
        self.ready = False

    async def start(self):
        # Start the Peer Service and Discovery service
        asyncio.ensure_future(
            self.peer_service.start()
        )

        if self.wallet.verifying_key().hex() in self.mn_to_find:
            asyncio.ensure_future(
                self.discovery_server.serve()
            )

        discovery_bootnode_ids = [services.SocketStruct(services.Protocols.TCP, ip, self.discovery_port)
                                     for ip in self.bootnodes]

        # Discover our bootnodes
        await self.discover_bootnodes(discovery_bootnode_ids)

        peer_service_bootnode_ids = [services.SocketStruct(services.Protocols.TCP, ip, self.peer_service_port)
                                     for ip in self.bootnodes]

        log.info('Peers now: {}'.format(self.bootnodes))

        # Wait for the quorum to resolve
        await self.wait_for_quorum(
            self.initial_mn_quorum,
            self.initial_del_quorum,
            self.mn_to_find,
            self.del_to_find,
            peer_service_bootnode_ids
        )

        log.success('Network is ready.')

        self.ready = True

        ready_msg = json.dumps({'event': 'service_status', 'status': 'ready'}, cls=services.SocketEncoder).encode()

        await self.peer_service.event_publisher.send(ready_msg)

        log.success('Sent ready signal.')
        return

    async def discover_bootnodes(self, nodes):
        responses = await discovery.discover_nodes(nodes, pepper=PEPPER.encode(),
                                                   ctx=self.ctx, timeout=3000)

        log.info(responses)

        for ip, vk in responses.items():
            self.table.peers[vk] = ip  # Should be stripped of port and tcp

        if not self.discovery_server.running:
            asyncio.ensure_future(self.discovery_server.serve())

        # Ping everyone discovered that you've joined

        current_nodes = deepcopy(self.table.peers)
        for vk, ip in current_nodes.items():
            join_message = ['join', (self.wallet.verifying_key().hex(), self.ip)]
            join_message = json.dumps(join_message, cls=services.SocketEncoder).encode()

            peer = services.SocketStruct(services.Protocols.TCP, ip, DHT_PORT)
            await services.get(peer, msg=join_message, ctx=self.ctx, timeout=3000)

    async def wait_for_quorum(self, masternode_quorum_required: int,
                                    delegate_quorum_required: int,
                                    masternodes_to_find: list,
                                    delegates_to_find: list,
                                    initial_peers: list,
                                    ):

        results = None


        # Crawl while there are still nodes needed in our quorum
        while masternode_quorum_required > 0 or delegate_quorum_required > 0:
            # Create task lists
            log.info('Need {} MNs and {} DELs to begin...'.format(
                masternode_quorum_required,
                delegate_quorum_required
            ))

            master_crawl = [self.find_node(client_address=random.choice(initial_peers),
                            vk_to_find=vk, retries=3) for vk in masternodes_to_find]

            delegate_crawl = [self.find_node(client_address=random.choice(initial_peers),
                              vk_to_find=vk, retries=3) for vk in delegates_to_find]

            # Crawl for both node types
            crawl = asyncio.gather(*master_crawl, *delegate_crawl)

            results = await crawl

            # Split the result list
            masters_got = results[:len(masternodes_to_find)]
            delegates_got = results[len(masternodes_to_find):]

            masternode_quorum_required = self.updated_peers_and_crawl(masters_got,
                                                                      masternodes_to_find,
                                                                      masternode_quorum_required)

            delegate_quorum_required = self.updated_peers_and_crawl(delegates_got,
                                                                    delegates_to_find,
                                                                    delegate_quorum_required)

        # At this point, start the discovery server if it's not already running because you are a masternode.

        log.success('Quorum reached! Begin protocol services...')

        return results

    def updated_peers_and_crawl(self, node_results, all_nodes, current_quorum):
        nodes = {}

        # Pack successful requests into a dictionary
        for node in node_results:
            if node is not None:
                nodes.update(node)

        # Update the peer table with the _new nodes
        self.table.peers.update(nodes)

        # Remove the nodes from the all_nodes list. Don't need to query them again
        for vk, _ in nodes.items():
            all_nodes.remove(vk)

        # Return the number of nodes needed left
        return current_quorum - len(nodes)

    # Returns raw IP string for a node: 127.0.0.1 etc. Transform this into whatever you want.
    async def find_node(self, client_address: services.SocketStruct=None, vk_to_find=None, retries=3):
        # Search locally if this is the case
        if str(client_address) == str(self.peer_service_address) or \
                vk_to_find == self.wallet.verifying_key().hex() or \
                client_address is None:
            response = self.table.find(vk_to_find)

        # Otherwise, send out a network request
        else:
            find_message = ['find', vk_to_find]
            find_message = json.dumps(find_message, cls=services.SocketEncoder).encode()
            response = await services.get(client_address, msg=find_message, ctx=self.ctx, timeout=3000)

            if response is None:
                return None

            response = json.loads(response.decode())

        if response.get(vk_to_find) is not None:
            return response

        if retries <= 1:
            return None

        # Recursive crawl goes 'retries' levels deep
        for vk, ip in response.items():
            return await self.find_node(services.SocketStruct(services.Protocols.TCP, ip, DHT_PORT), vk_to_find, retries=retries-1)



import glob
import os
from contracting.client import ContractingClient
from cilantro_ee.core.logger.base import get_logger

log = get_logger("sync")

def contract_name_from_file_path(p: str) -> str:
    directories = p.split('/')
    filename = directories[-1]

    name_and_file_extention = filename.split('.')
    name = name_and_file_extention[0]

    return name


def contracts_for_directory(path, extension, directory=os.path.dirname(__file__)):
    dir_path = os.path.join(directory, path) + '/' + extension
    contracts = glob.glob(dir_path)
    return contracts


def sync_genesis_contracts(genesis_path: str='genesis',
                           extension: str='*.s.py',
                           exclude=['vkbook'],
                           directory=os.path.dirname(__file__)):

    # Direct database writing of all contract files in the 'genesis' folder
    # direct_contracts = contracts_for_directory(direct_path, extension)
    # explicitly submit the submission contract
    submission_file = directory + '/submission.s.py'
    client = ContractingClient(submission_filename=submission_file)

    genesis_contracts = contracts_for_directory(genesis_path, extension, directory=directory)

    for contract in genesis_contracts:
        name = contract_name_from_file_path(contract)
        if name in exclude:
            continue

        if client.raw_driver.get_contract(name) is None:
            with open(contract) as f:
                code = f.read()

            client.submit(code, name=name)


def submit_contract_with_construction_args(name, directory=os.path.dirname(__file__), args={}):
    file = directory + '/genesis/{}.s.py'.format(name)

    submission_file = os.path.dirname(__file__) + '/submission.s.py'
    client = ContractingClient(submission_filename=submission_file)

    with open(file) as f:
        code = f.read()
        # log.debug('code {}'.format(code))
        # log.debug('name {}'.format(name))
        # log.debug('args {}'.format(args))
        client.submit(code, name=name, constructor_args=args)

    client.raw_driver.commit()



"""
Utility class for hashing thangs
"""
import hashlib
from cilantro_ee.core.messages.capnp_impl.capnp_impl import pack, unpack


class Hasher:
    """
    Just an enum of hashing algs (print property hashlib.algorithms_available for full list)
    """
    class Alg:
        MD5 = 'md5'
        SHA3_256 = 'sha3_256'
        SHA1 = 'sha1'
        RIPE = 'sha256'
        SHAKE_128 = 'shake_128'
        SHAKE_256 = 'shake_256'

    DEFAULT_ALG = Alg.SHA3_256

    class Test:
        SOME_PROP = 1

    @staticmethod
    def _cast_to_bytes(data) -> bytes:
        """
        Attempts to auto-cast the data to bytes, raising an error if not possible
        :param data: The data to attempt to cast to bytes
        :return: Bytes
        :raises: An assertion if data is a non-trivial type that could not be casted to bytes
        """
        from cilantro_ee.utils import int_to_bytes

        assert data is not None, "Cannot hash a None type!"

        t = type(data)

        if t is str:
            data = data.encode()
        elif t is int:
            data = pack(data)

        assert type(data) is bytes, "Unable to cast data of original type {} into bytes".format(t)

        return data

    @staticmethod
    def _read_hasher(hasher, return_bytes=False, digest_len=0):
        if digest_len > 0:
            binary = hasher.digest(digest_len)
        else:
            binary = hasher.digest()

        if return_bytes:
            return binary
        else:
            return binary.hex()

    @staticmethod
    def hash(data, algorithm=DEFAULT_ALG, return_bytes=False, digest_len: int=0) -> bytes or str:
        """
        Attempts to automatically cast the data to bytes, and hash it. If digest_len is specified, shake_256 will be
        used (unless algorithm is set to shake_128) to compute a variable size digest.
        :param data: The data to be hashes. This method will do its best to infer its type and cast it to bytes, but
        to be sure it will work you can pass in bytes explicity
        :param algorithm: The algorithm to use (a property of Hasher.Alg)
        :param return_bytes: If true, returns the hash as bytes. Otherwise, returns a hex string
        :param digest_len: The optional length of the digest (as an int). Max is 256.
         If this is specified, algorithm must be either Alg.SHAKE_256 or Alg.SHAKE_128
        :return: A string containing the hex digest of the resulting hash. If return_bytes is True, then this hex string
        is returned in binary format
        """
        data = Hasher._cast_to_bytes(data)

        if digest_len > 0:
            if algorithm != Hasher.DEFAULT_ALG:
                assert algorithm in (Hasher.Alg.SHAKE_128, Hasher.Alg.SHAKE_256), \
                    "If digest_len is set, algorithm must be Alg.SHAKE_128 or Alg.SHAKE_256"
            else:
                algorithm = Hasher.Alg.SHAKE_256

            assert digest_len <= 256//8, "digest_len must be less than or equal to 32"

            if digest_len > 128//8:
                assert algorithm == Hasher.Alg.SHAKE_256, "digest_len greater than 16 must use SHAKE_256"

        h = hashlib.new(algorithm)
        h.update(data)

        return Hasher._read_hasher(h, return_bytes=return_bytes, digest_len=digest_len)

    @staticmethod
    def hash_iterable(iterable, algorithm=DEFAULT_ALG, return_bytes=False) -> bytes or str:
        """
        Hashes an iterable by casting all its elements to bytes (if necessary), concatenating them, and then hashing
        the resulting binary
        :param iterable: An iterable to hash. Elements can be homo/heterogeneous.
        :param algorithm: The algorithm to use (a property of Hasher.Alg
        :param return_bytes: If true, the hash will be returned as bytes. Otherwise, a hex string will be returned
        :return: The resulting hash in bytes or as a str
        """
        # Concatenate individual binary datums
        data = b''
        for i in iterable:
            data += Hasher._cast_to_bytes(i)

        hasher = hashlib.new(algorithm)
        hasher.update(data)

        return Hasher._read_hasher(hasher, return_bytes=return_bytes)



from cilantro_ee.messages.transaction.publish import *
from cilantro_ee.messages.transaction.contract import *

from cilantro_ee.core.logger import get_logger
from cilantro_ee.utils.test.utils import *
from cilantro_ee.utils.test.wallets import ALL_WALLETS
import os, requests, time, random, asyncio, secrets, math, json
from typing import List
from decimal import Decimal


CURRENCY_CONTRACT_NAME = 'currency'



class God:
    # For MP tests  TODO i dont think we need this  --davis
    node_map = None
    testers = []

    log = get_logger("GOD")

    mn_urls = []
    _current_mn_idx = 0

    def __init__(self):
        raise NotImplementedError("Use only class method on God")

    @classmethod
    def _default_gen_func(cls):
        return cls.random_contract_tx

    @classmethod
    def create_currency_tx(cls, sender: tuple, receiver: tuple, amount: int, stamps=10000, nonce=None):
        if type(receiver) is tuple:
            receiver = receiver[1]

        return ContractTransactionBuilder.create_currency_tx(sender[0], receiver, amount, stamps=stamps, nonce=nonce)

    @classmethod
    def create_contract_tx(cls, sender_sk: str, contract_name: str, func_name: str, stamps=10000,
                           nonce=None, **kwargs) -> ContractTransaction:
        if nonce is None:
            nonce = wallet.get_vk(sender_sk) + ':' + 'A'*64
        return ContractTransaction.create(sender_sk, stamps, contract_name, func_name, nonce, kwargs)

    @classmethod
    def send_currency_contract(cls, sender: tuple, receiver: tuple, amount: int):
        tx = cls.create_currency_tx(sender, receiver, amount)
        return cls.send_tx(tx)

    @classmethod
    def send_tx(cls, tx):
        assert len(cls.mn_urls) > 0, "mn_urls must be set to send_tx! "
        mn_url = cls._get_mn_url()
        data = TransactionContainer.create(tx).serialize()
        try:
            r = requests.post(mn_url, data=data, verify=False)
            cls.log.spam("POST request to MN at URL {} has status code: {}".format(mn_url, r.status_code))
            return r
        except Exception as e:
            cls.log.warning("Error attempt to send transaction to Masternode at URL {}\nerror={}".format(mn_url, e))
            return None

    @classmethod
    async def async_send_txs(cls, txs: List[TransactionBase], session):
        async def _send(url, data):
            async with session.post(url, data=data) as resp:
                return await resp.json()

        tx_containers = [TransactionContainer.create(tx).serialize() for tx in txs]
        futs = []
        for data in tx_containers:
            url = cls._get_mn_url()
            futs.append(_send(url, data))
        await asyncio.gather(*futs)

    @classmethod
    def pump_it(cls, rate: int, gen_func=None, use_poisson=True, sleep_sometimes=False, active_bounds=(120, 240),
                sleep_bounds=(20, 60), pump_wait=0):
        """
        Pump random transactions from random users to Masternode's REST endpoint at an average rate of 'rate'
        transactions per second. This func blocks.
        """
        if pump_wait > 0:
            cls.log.important("Pumper sleeping {} seconds before starting...".format(pump_wait))
            time.sleep(pump_wait)

        if not gen_func:
            gen_func = cls._default_gen_func()

        if use_poisson:
            from scipy.stats import poisson, expon
            rvs_func = lambda: expon.rvs(rate)/rate - 1
        else:
            rvs_func = lambda: 1/rate

        assert callable(gen_func), "Expected a callable for 'gen_func' but got {}".format(gen_func)

        cls.log.important3("Starting to pump transactions at an average of {} transactions per second".format(rate))
        cls.log.test("Using generator func {}, use_possion={}, sleep_sometimes={}, active_bounds={}, sleep_bounds={}"
                     .format(gen_func, use_poisson, sleep_sometimes, active_bounds, sleep_bounds))

        time_since_last_sleep = 0
        next_sleep = random.randint(active_bounds[0], active_bounds[1])
        if sleep_sometimes:
            cls.log.important3("Next sleep will be in {}s".format(next_sleep))

        while True:
            wait = rvs_func()
            # cls.log.spam("Sending next transaction in {} seconds".format(wait))
            time.sleep(wait)
            time_since_last_sleep += wait

            tx = gen_func()
            # cls.log.spam("sending transaction {}".format(tx))
            cls.send_tx(tx)

            if sleep_sometimes and time_since_last_sleep >= next_sleep:
                sleep_time = random.randint(sleep_bounds[0], sleep_bounds[1])
                cls.log.important3("Sleeping for {}s before pumping more...")
                time.sleep(sleep_time)

                time_since_last_sleep = 0
                next_sleep = random.randint(active_bounds[0], active_bounds[1])
                cls.log.important3("Done sleeping. Continuing the pump, and triggering next sleep in {}s".format(next_sleep))

    @classmethod
    async def dump_it(cls, session, volume: int, delay: int=0, gen_func=None):
        """ Dump it fast. """
        assert volume > 0, "You must dump at least 1 transaction silly"

        if not gen_func:
            gen_func = cls._default_gen_func()

        gen_start_time = time.time()
        cls.log.important2("Generating {} transactions to dump...".format(volume))
        txs = [gen_func() for _ in range(volume)]
        cls.log.important2("Done generating transactions.")

        delay -= int(time.time() - gen_start_time)
        countdown(delay, "Waiting for an additional {} seconds before dumping...", cls.log, status_update_freq=8)

        start = time.time()
        cls.log.info("Dumping {} transactions...".format(len(txs)))
        await cls.async_send_txs(txs, session)
        # for tx in txs:
        #     cls.send_tx(tx)
        cls.log.success("Done dumping {} transactions in {} seconds".format(len(txs), round(time.time() - start, 3)))

    @classmethod
    def request_nonce(cls, vk):
        mn_url = cls._get_mn_url() + '/nonce'
        try:
            r = requests.get(mn_url, json={'verifyingKey': vk})
            cls.log.debugv("GET request to MN at URL {} has status code: {}".format(mn_url, r.status_code))
            return r.json()

        except Exception as e:
            cls.log.warning("Error attempt to send transaction to Masternode at URL {}\nerror={}".format(mn_url, e))
            return 'error: {}'.format(e)

    @classmethod
    def random_contract_tx(cls):
        sender, receiver = random.sample(ALL_WALLETS, 2)
        amount = random.randint(1, 100)

        return cls.create_currency_tx(sender=sender, receiver=receiver, amount=amount)

    @classmethod
    def get_random_mn_url(cls):
        return random.choice(cls.mn_urls)

    @classmethod
    def _parse_reply(cls, req, req_type='json'):
        if req.status_code != 200:
            cls.log.spam("Got status code {} from request {}".format(req.status_code, req))
            return None

        if req_type == 'json':
            return req.json()
        else:
            raise Exception("Unknown request type {}".format(req_type))

    @classmethod
    def _check_eq_replies(cls, replies: dict):
        vals = list(replies.values())
        if all(x == vals[0] for x in vals):
            return vals[0]
        else:
            cls.log.warning("Masternodes had inconsistent replies for GET requests ... possibile state"
                            " corruption!?\nReplies: {}".format(replies))
            return None

    @classmethod
    def get_from_mn_api(cls, query_str, enforce_consistency=True, req_type='json'):
        if not enforce_consistency:
            return cls._parse_reply(requests.get("{}/{}".format(cls.get_random_mn_url(), query_str)))

        replies = {}
        for mn_url in cls.mn_urls:
            req_url = "{}/{}".format(mn_url, query_str)
            replies[req_url] = cls._parse_reply(requests.get(req_url))
        return cls._check_eq_replies(replies)

    @classmethod
    async def async_get_from_mn_api(cls, query_str, session, enforce_consistency=True, req_type='json'):
        async def fetch_url(url):
            async with session.get(url) as resp:
                if req_type == 'json':
                    return await resp.json()
                elif req_type == 'text':
                    return await resp.text()
                else:
                    raise Exception("Unknown req_type {}".format(req_type))

        if not enforce_consistency:
            return await fetch_url("{}/{}".format(cls.mn_urls[0], query_str))

        replies = {}
        for mn_url in cls.mn_urls:
            req_url = "{}/{}".format(mn_url, query_str)
            replies[req_url] = fetch_url(req_url)

        results = dict(zip(replies.keys(), await asyncio.gather(*list(replies.values()))))
        return cls._check_eq_replies(results)

    @classmethod
    def submit_contract(cls, code: str, name: str, sk: str, vk: str, stamps=10**6):
        tx = PublishTransaction.create(contract_code=code,
                                       contract_name=name,
                                       sender_sk=sk,
                                       nonce=vk + secrets.token_hex(32),
                                       stamps_supplied=stamps)
        return cls.send_tx(tx)

    @classmethod
    def get_contract_names(cls) -> List[str]:
        r = cls.get_from_mn_api('contracts')
        if r is not None:
            assert 'contracts' in r, "Expected key 'contracts' to be in dict returned by /contracts endpoint. " \
                                     "Instead got {}".format(r)
            return r['contracts']
        else:
            return None

    @classmethod
    def get_contract_meta(cls, contract_name: str):
        return cls.get_from_mn_api('contracts/{}'.format(contract_name))

    @classmethod
    def get_contract_resources(cls, contract_name: str):
        return cls.get_from_mn_api('contracts/{}/resources'.format(contract_name))

    @classmethod
    def get_contract_methods(cls, contract_name: str):
        return cls.get_from_mn_api('contracts/{}/methods'.format(contract_name))

    @classmethod
    async def get_balances(cls, session, vks: list, contract_name=CURRENCY_CONTRACT_NAME) -> dict:
        async def _get_balance(vk, contract_name):
            req_url = "contracts/{}/balances/{}".format(contract_name, vk)
            return cls._process_balance_json(await God.async_get_from_mn_api(req_url, session))

        balances, fetched = {}, {}

        for vk in vks:
            balances[vk] = _get_balance(vk, contract_name=contract_name)
        results = dict(zip(balances.keys(), await asyncio.gather(*list(balances.values()))))

        for k in results:
            if results[k] is not None:
                fetched[k] = results[k]

        return fetched

    @classmethod
    def wait_for_mns_online(cls, timeout=300):
        cls.log.notice("Waiting for masternodes to come online...")
        elapsed = 0
        while True:
            try:
                cls.log.info("Checking if masternodes are online")
                val = God.get_from_mn_api('ping')
                if val is not None:
                    cls.log.notice("All masternodes online.")
                    return
            except Exception as e:
                cls.log.warning("Error checking if MNs online: {}".format(e))

            if elapsed > timeout:
                raise Exception("Timeout of {} reached waiting for masternodes to be online")

            elapsed += 5
            time.sleep(5)

    @classmethod
    async def get_balance(cls, vk, contract_name=CURRENCY_CONTRACT_NAME) -> int or None:
        req_url = "contracts/{}/balances/{}".format(contract_name, vk)
        return cls._process_balance_json(God.get_from_mn_api(req_url))

    @classmethod
    def _process_balance_json(cls, d: dict) -> int or None:
        try:
            if d:
                assert 'value' in d, "Expected key 'value' to be in reply json {}".format(d)
                return Decimal(d['value'])
            else:
                return None
        except Exception as e:
            log.critical("Got error processing balance json {} ... error:\n {}".format(d, e))


    @classmethod
    def _get_mn_url(cls):
        assert len(cls.mn_urls) > 0, "mn_urls not set!"
        mn_url = cls.mn_urls[cls._current_mn_idx]
        cls._current_mn_idx = (cls._current_mn_idx + 1) % len(cls.mn_urls)
        return mn_url




import os


MASTER_DB = 0

DATA_DIR = os.getenv('DATADIR')

if DATA_DIR is None:
    DATA_DIR = '/usr/local/db/cilantro_ee'


MONGO_DIR = DATA_DIR + '/mongo'
MONGO_LOG_PATH = MONGO_DIR + '/logs/mongo.log'


def config_mongo_dir():
    try:
        os.makedirs(os.path.dirname(MONGO_LOG_PATH), exist_ok=True)
        with open(MONGO_LOG_PATH, 'w+') as f:
            print('Mongo log file created at path {}'.format(MONGO_LOG_PATH))
    except Exception as e:
        print("Error creating mongo log file at path {}. Error -- {}".format(MONGO_LOG_PATH, e))



# def wrap_func(fn, *args, **kwargs):
#     def wrapper():
#         return fn(*args, **kwargs)
#     return wrapper
#
#
# def log_create(name, vk, ip):
#     from cilantro_ee.core.logger import get_logger
#     log = get_logger("{}Builder".format(name))
#     delim = '=' * 64
#     msg = '\n' + delim + '\n' + 'Creating {} with:\nvk={}\nip={}\n'.format(name, vk, ip) + delim
#     log.test(msg)
#
#
# def run_mn(slot_num=None, sk=None, log_lvl=11, reset_db=False, nonce_enabled=True):
#     assert slot_num is not None or sk is not None, "SK or slot num must be provided"
#
#     # Due to some Sanic BS, we cannot set the log level between [1,10]
#     assert log_lvl not in range(2, 11), "Due to a Sanic logging bug, Masternode cant set log lvl in the range (1,10]"
#
#     from cilantro_ee.constants.conf import CilantroConf
#     from cilantro_ee.core.logger import get_logger, overwrite_logger_level
#     from cilantro_ee.nodes.factory import NodeFactory
#     from cilantro_ee.utils.test.node_runner import log_create
#     from cilantro_ee.core.crypto import wallet
#
#     overwrite_logger_level(log_lvl)
#     vk = wallet.get_vk(sk)
#
#     ip = CilantroConf.HOST_IP
#     log_create("Masternode", vk, ip)
#     NodeFactory.run_masternode(ip=ip, signing_key=sk, reset_db=reset_db)
#
#
# def run_witness(slot_num=None, sk=None, log_lvl=11, reset_db=False):
#     assert slot_num is not None or sk is not None, "SK or slot num must be provided"
#     from cilantro_ee.core.logger import get_logger, overwrite_logger_level
#     from cilantro_ee.nodes.factory import NodeFactory
#     from cilantro_ee.constants.testnet import TESTNET_WITNESSES
#     from cilantro_ee.constants.conf import CilantroConf
#     from cilantro_ee.utils.test.node_runner import log_create
#     from cilantro_ee.core.crypto import wallet
#
#     overwrite_logger_level(log_lvl)
#
#     vk = wallet.get_vk(sk)
#
#     ip = CilantroConf.HOST_IP
#     log_create("Witness", vk, ip)
#     NodeFactory.run_witness(ip=ip, signing_key=sk, reset_db=reset_db)
#
#
# def run_delegate(slot_num=None, sk=None, log_lvl=11, seneca_log_lvl=11, bad_actor=False, reset_db=False, bad_sb_set={1}, num_succ_sbs=3):
#     assert slot_num is not None or sk is not None, "SK or slot num must be provided"
#     from cilantro_ee.constants.conf import CilantroConf
#     import os
#     if bad_actor:
#         os.environ["BAD_ACTOR"] = '1'
#         os.environ["SB_IDX_FAIL"] = ','.join((str(i) for i in bad_sb_set))
#         os.environ["NUM_SUCC_SBS"] = str(num_succ_sbs)
#
#     from cilantro_ee.core.logger import get_logger, overwrite_logger_level
#     from eneca.libs.logger import overwrite_logger_level as sen_overwrite_log
#     from cilantro_ee.nodes.factory import NodeFactory
#     from cilantro_ee.utils.test.node_runner import log_create
#     from cilantro_ee.core.crypto import wallet
#
#     overwrite_logger_level(log_lvl)
#     sen_overwrite_log(seneca_log_lvl)
#
#     vk = wallet.get_vk(sk)
#
#     ip = CilantroConf.HOST_IP
#     log_create("Delegate", vk, ip)
#     NodeFactory.run_delegate(ip=ip, signing_key=sk, reset_db=reset_db)
#
#
# def dump_it(volume, delay=0):
#     from cilantro_ee.utils.test.god import God
#     from cilantro_ee.core.logger import get_logger, overwrite_logger_level
#     import logging
#
#     overwrite_logger_level(logging.WARNING)
#     God.dump_it(volume=volume, delay=delay)
#
#
# def pump_it(*args, **kwargs):
#     from cilantro_ee.utils.test.god import God
#     from cilantro_ee.core.logger import get_logger, overwrite_logger_level
#     import logging, time
#
#     overwrite_logger_level(logging.WARNING)
#
#     log = get_logger("Pumper")
#     log.important("Starting the pump..")
#     God.pump_it(*args, **kwargs)



from unittest import TestCase
from cilantro_ee.services.storage.master import CilantroStorageDriver
from cilantro_ee.core.crypto import wallet


class TestCilantroStorageDriver(TestCase):
    def setUp(self):
        sk, vk = wallet.new()
        self.db = CilantroStorageDriver(key=sk)

    def tearDown(self):
        self.db.drop_collections()

    def test_init(self):
        self.assertIsNotNone(self.db)



from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('2-2-2.json')

from cilantro_ee.messages.base.base import MessageBase
from cilantro_ee.messages.envelope.envelope import Envelope
from cilantro_ee.messages.consensus.sub_block_contender import SubBlockContender, SubBlockContenderBuilder
from cilantro_ee.messages.consensus.merkle_signature import MerkleSignature, build_test_merkle_sig
from cilantro_ee.messages.transaction.data import TransactionDataBuilder
from cilantro_ee.constants.testnet import TESTNET_MASTERNODES, TESTNET_DELEGATES
from cilantro_ee.core.containers.merkle_tree import MerkleTree
from unittest import TestCase
from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook
import unittest


TEST_SK = TESTNET_MASTERNODES[0]['sk']
TEST_VK = TESTNET_MASTERNODES[0]['vk']
DEL_SK = TESTNET_DELEGATES[0]['sk']
DEL_VK = TESTNET_DELEGATES[0]['vk']


class TestSubBlockContender(TestCase):

    def setUp(self):
        PhoneBook = VKBook(delegates=[DEL_VK], masternodes=[TEST_VK])

    def test_builder(self):
        sbc = SubBlockContenderBuilder.create()
        self.assertEqual(sbc, SubBlockContender.from_bytes(sbc.serialize()))

    def test_create(self):
        txs = [TransactionDataBuilder.create_random_tx() for i in range(5)]
        raw_txs = [tx.serialize() for tx in txs]
        tree = MerkleTree.from_raw_transactions(raw_txs)

        input_hash = 'B' * 64  # in reality this would be the env hash. we can just make something up
        signature = build_test_merkle_sig(msg=tree.root, sk=DEL_SK, vk=DEL_VK)

        sbc1 = SubBlockContender.create(result_hash=tree.root_as_hex, input_hash=input_hash, merkle_leaves=tree.leaves,
                                        signature=signature, transactions=txs, sub_block_index=0, prev_block_hash='0'*64)
        sbc2 = SubBlockContender.create(result_hash=tree.root_as_hex, input_hash=input_hash, merkle_leaves=tree.leaves,
                                        signature=signature, transactions=txs, sub_block_index=0, prev_block_hash='0'*64)
        self.assertFalse(sbc1.is_empty)
        self.assertFalse(sbc2.is_empty)
        self.assertEqual(sbc1, sbc2)

    def test_serialize_deserialize(self):
        txs = [TransactionDataBuilder.create_random_tx() for i in range(5)]
        raw_txs = [tx.serialize() for tx in txs]
        tree = MerkleTree.from_raw_transactions(raw_txs)

        input_hash = 'B' * 64  # in reality this would be the env hash. we can just make something up
        signature = build_test_merkle_sig(msg=tree.root, sk=DEL_SK, vk=DEL_VK)

        sbc = SubBlockContender.create(result_hash=tree.root_as_hex, input_hash=input_hash, merkle_leaves=tree.leaves,
                                       signature=signature, transactions=txs, sub_block_index=0, prev_block_hash='0'*64)
        clone = SubBlockContender.from_bytes(sbc.serialize())

        self.assertEqual(clone, sbc)

    def test_empty_sub_block_contender(self):
        input_hash = 'B' * 64  # in reality this would be the env hash. we can just make something up
        signature = build_test_merkle_sig(msg=bytes.fromhex(input_hash), sk=DEL_SK, vk=DEL_VK)

        sbc = SubBlockContender.create_empty_sublock(input_hash=input_hash, signature=signature, sub_block_index=0,
                                                     prev_block_hash='0' * 64)

        self.assertTrue(sbc.is_empty)

    def test_sbc_handshake(self):
        txs = [TransactionDataBuilder.create_random_tx() for i in range(5)]
        raw_txs = [tx.serialize() for tx in txs]
        tree = MerkleTree.from_raw_transactions(raw_txs)

        input_hash = 'B' * 64  # in reality this would be the env hash. we can just make something up
        signature = build_test_merkle_sig(msg=tree.root, sk=DEL_SK, vk=DEL_VK)

        sbc1 = SubBlockContender.create(result_hash=tree.root_as_hex, input_hash=input_hash, merkle_leaves=tree.leaves,
                                       signature=signature, transactions=txs, sub_block_index=0, prev_block_hash='0'*64)
        self.assertFalse(sbc1.is_empty)
        #msg_type = MessageBase.registry[type(sbc1)]
        sbc2 = SubBlockContender.from_bytes(sbc1.serialize())
        self.assertEqual(sbc1, sbc2)

        msg2 = Envelope.create_from_message(message=sbc2.serialize(), signing_key=DEL_SK, verifying_key=DEL_VK)
        es = msg2.serialize()
        env = Envelope.from_bytes(es)
        sbc = env.message
        if not sbc.signature.verify(bytes.fromhex(sbc.result_hash)):
            print("This SubBlockContender does not have a correct signature!")
        if len(sbc.merkle_leaves) > 0:
            print("This SubBlockContender have num of merkle leaves!")
            print(len(sbc.merkle_leaves))
            if MerkleTree.verify_tree_from_str(sbc.merkle_leaves, root=sbc.result_hash):
                for tx in sbc.transactions:
                    if not tx.hash in sbc.merkle_leaves:
                        print('Received malicious transactions that does not match any merkle leaves!')
                        break
                print('This SubBlockContender is valid!')
            else:
                print('This SubblockContender is INVALID!')
        else:
            print('This SubBlockContender is empty.')




if __name__ == '__main__':
    unittest.main()



import time
import asyncio
from cilantro_ee.core.logger import get_logger
from cilantro_ee.constants.zmq_filters import *
from cilantro_ee.core.sockets.lsocket import LSocketBase
from cilantro_ee.services.storage.vkbook import PhoneBook
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.services.storage.master import CilantroStorageDriver
from cilantro_ee.services.storage.master import MasterStorage
from cilantro_ee.contracts.sync import sync_genesis_contracts

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message

from cilantro_ee.core.nonces import NonceManager

IDX_REPLY_TIMEOUT = 20
TIMEOUT_CHECK_INTERVAL = 1


class CatchupManager:
    def __init__(self, verifying_key: str, signing_key: str, pub_socket, router_socket: LSocketBase, store_full_blocks=True):
        """

        :param verifying_key: host vk
        :param pub_socket:
        :param router_socket:
        :param store_full_blocks: Master node uses this flag to indicate block storage
        """
        self.log = get_logger("CatchupManager")

        # infra input
        self.pub, self.router = pub_socket, router_socket
        self.verifying_key = verifying_key
        self.signing_key = signing_key
        self.store_full_blocks = store_full_blocks

        # self.blocks = CilantroStorageDriver(key=self.signing_key)

        self.state = MetaDataStorage()
        self.nonce_manager = NonceManager()

        # catchup state
        self.is_caught_up = False
        self.timeout_catchup = time.time()      # 10 sec time we will wait for 2/3rd MN to respond
        self.node_idx_reply_set = set()  # num of master responded to catch up req

        # main list to process
        self.block_delta_list = []       # list of mn_index dict to process

        # received full block could be out of order
        self.rcv_block_dict = {}                 # DS stores any Out of order received blocks

        # loop to schedule timeouts
        self.timeout_fut = None

        self.my_quorum = PhoneBook.masternode_quorum_min

        # masternode should make sure redis and mongo are in sync
        if store_full_blocks:
            self.my_quorum -= 1
            self.update_state()

        self.curr_hash = self.state.latest_block_hash
        self.curr_num = self.state.latest_block_num

        self.target_blk_num = self.curr_num
        self.awaited_blknum = self.curr_num

        # DEBUG -- TODO DELETE
        self.log.test("CatchupManager VKBook MN's: {}".format(PhoneBook.masternodes))
        self.log.test("CatchupManager VKBook Delegates's: {}".format(PhoneBook.delegates))
        # END DEBUG

        self.nonce_manager = NonceManager(driver=self.state)

        if self.store_full_blocks:
            self.driver = CilantroStorageDriver(key=self.signing_key)
        else:
            self.driver = None

    def update_state(self):
        """
        Sync block and state DB if either is out of sync.
        :return:
        """
        self.driver = CilantroStorageDriver(key=self.signing_key)

        last_block = self.driver.get_last_n(1, MasterStorage.INDEX)[0]

        db_latest_blk_num = last_block.get('blockNum')

        latest_state_num = self.state.latest_block_num
        if db_latest_blk_num < latest_state_num:
            # TODO - assert and quit
            self.log.fatal("Block DB block - {} is behind StateDriver block - {}. Cannot handle"
                           .format(db_latest_blk_num, latest_state_num))
            # we need to rebuild state from scratch
            latest_state_num = 0
            self._reset_state()

        if db_latest_blk_num > latest_state_num:
            self.log.info("StateDriver block num {} is behind DB block num {}".format(latest_state_num, db_latest_blk_num))
            while latest_state_num < db_latest_blk_num:
                latest_state_num = latest_state_num + 1
                # TODO get nth full block wont work for now in distributed storage
                blk_dict = self.driver.get_block(latest_state_num)

                if '_id' in blk_dict:
                    del blk_dict['_id']

                block = Message.get_message(MessageType.BLOCK_DATA, **blk_dict)
                self.state.update_with_block(block=block)

        # Reinitialize the latest nonce. This should probably be abstracted into a seperate class at a later date
        blk_dict = self.driver.get_block(latest_state_num)
        sbs = blk_dict.get('subBlocks')

        nonces = {}

        # raghu todo - need to revisit this? why do we serialize sub-blocks only, but not block
        if sbs is not None:
            for raw_sb in blk_dict['subBlocks']:
                subblock = Message.unpack_message_internal(MessageType.SUBBLOCK, raw_sb)
                self.log.info('Block: {}'.format(subblock))
                for tx in subblock.transactions:
                    self.nonce_manager.update_nonce_hash(nonce_hash=nonces, tx_payload=tx.transaction.payload)
                    self.state.set_transaction_data(tx=tx)

        self.nonce_manager.commit_nonces(nonce_hash=nonces)
        self.nonce_manager.delete_pending_nonces()

        self.log.info("Verify StateDriver num {} StorageDriver num {}".format(latest_state_num, db_latest_blk_num))

    # should be called only once per node after bootup is done
    def run_catchup(self, ignore=False):
        self.log.info("-----RUN CATCHUP-----")
        # check if catch up is already running
        if ignore and self.is_catchup_done():
            self.log.warning("Already caught up. Ignoring to run it again.")
            return

        if self.my_quorum == 0:    # only one master available
            self.is_caught_up = True
            return

        # first reset state variables
        self.node_idx_reply_set.clear()
        self.is_caught_up = False
        # self.curr_hash, self.curr_num = StateDriver.get_latest_block_info()
        # self.target_blk_num = self.curr_num
        # self.awaited_blknum = None

        # starting phase I
        self.timeout_catchup = time.time()
        self.send_block_idx_req()

        self._reset_timeout_fut()
        # first time wait longer than usual
        time.sleep(3 * TIMEOUT_CHECK_INTERVAL)
        self.timeout_fut = asyncio.ensure_future(self._check_timeout())
        self.log.important2("Running catchup!")

    def _reset_state(self):
        # only in a very rare case where mongo db is behind redis, this is called
        self.state.flush()
        sync_genesis_contracts()

    def _reset_timeout_fut(self):
        if self.timeout_fut:
            if not self.timeout_fut.done():
                # TODO not sure i need this try/execpt here --davis
                try: self.timeout_fut.cancel()
                except: pass
            self.timeout_fut = None

    async def _check_timeout(self):
        async def _timeout():
            elapsed = 0
            while elapsed < IDX_REPLY_TIMEOUT:
                elapsed += TIMEOUT_CHECK_INTERVAL
                await asyncio.sleep(TIMEOUT_CHECK_INTERVAL)

                if self._check_idx_reply_quorum() is True:
                    self.log.debugv("Quorum reached!")
                    return

            # If we have not returned from the loop and the this task has not been canceled, initiate a retry
            self.log.warning("Timeout of {} reached waiting for block idx replies! Resending BlockIndexRequest".format(IDX_REPLY_TIMEOUT))
            self.timeout_fut = None
            self.run_catchup(ignore=True)

        try:
            await _timeout()
        except asyncio.CancelledError as e:
            pass

    # Phase I start
    def send_block_idx_req(self):
        """
        Multi-casting BlockIndexRequests to all master nodes with current block hash
        :return:
        """
        self.log.info("Multi cast BlockIndexRequests to all MN with current block hash {}".format(self.curr_hash))

        msg_type, msg = Message.get_message_packed(
                                      MessageType.BLOCK_INDEX_REQUEST,
                                      blockHash=self.curr_hash,
                                      sender=self.verifying_key)

        self.pub.send_msg(BLOCK_IDX_REQ_FILTER.encode(), msg_type, msg)

    def _recv_block_idx_reply(self, sender_vk: str, reply):
        self.log.info('Got REPLY from {} as {}'.format(sender_vk, reply))
        """
        We expect to receive this message from all mn/dn
        :param sender_vk:
        :param reply:
        :return:
        """
        if sender_vk in self.node_idx_reply_set:
            return      # already processed

        if reply == b'':
            self.node_idx_reply_set.add(sender_vk)
            self.log.important("Received BlockIndexReply with no index info from masternode {}".format(sender_vk))
            return

        tmp_list = reply.get('indices')
        if len(tmp_list) > 1:
            assert tmp_list[0].get('blockNum') > tmp_list[-1].get('blockNum'), "ensure reply are in ascending order {}"\
                .format(tmp_list)
            tmp_list.reverse()

        self.log.debugv("tmp list -> {}".format(tmp_list))
        self.new_target_blk_num = tmp_list[-1].get('blockNum') if len(tmp_list) > 0 else 0
        new_blks = self.new_target_blk_num - self.target_blk_num

        if new_blks > 0:
            self.target_blk_num = self.new_target_blk_num
            update_list = tmp_list[-new_blks:]
            self.block_delta_list.extend(update_list)
            if self.awaited_blknum == self.curr_num:
                self.process_recv_idx()

        self.node_idx_reply_set.add(sender_vk)
        self.log.debugv("_new target block num {}\ntarget block num {}\ntemp list {}"
                        .format(self.new_target_blk_num, self.target_blk_num, tmp_list))

    def recv_block_idx_reply(self, sender_vk: str, reply):
        self._recv_block_idx_reply(sender_vk, reply.to_dict())
        # self.log.important2("RCV BIRp")
        return self.is_catchup_done()

    def _send_block_data_req(self, mn_vk, req_blk_num):
        self.log.info("Unicast BlockDateRequests to masternode owner with current block num {} key {}"
                      .format(req_blk_num, mn_vk))
        msg_type, msg = Message.get_message_packed(
                                      MessageType.BLOCK_DATA_REQUEST,
                                      blockNum=req_blk_num)
        self.router.send_msg(mn_vk, msg_type, msg)

    def _recv_block_data_reply(self, reply):
        # check if given block is older thn expected drop this reply
        # check if given blocknum grter thn current expected blk -> store temp
        # if given block needs to be stored update state/storage delete frm expected DT
        self.log.info('Recieved {}:'.format(reply))

        rcv_blk_num = reply.blockNum
        if rcv_blk_num <= self.curr_num:
            self.log.debug2("dropping already processed blk reply blk-{}:hash-{} ".format(reply.blockNum, reply.blockHash))
            return

        self.rcv_block_dict[rcv_blk_num] = reply
        # WHY IS AWAITED BLK NUM NONE HERE ???
        if rcv_blk_num > self.awaited_blknum:
            self.log.debug2("Got block num {}, still awaiting block num {}".format(rcv_blk_num, self.awaited_blknum))
            return

        if (rcv_blk_num == self.awaited_blknum):
            self.log.info('Got the block I needed!')
            self.curr_num = self.awaited_blknum
            self.update_received_block(block = reply)
            self.process_recv_idx()

    def recv_block_data_reply(self, block):
        self._recv_block_data_reply(block)
        return self.is_catchup_done()

    # MASTER ONLY CALL
    def recv_block_idx_req(self, request):
        """
        Receive BlockIndexRequests calls storage blocks to process req and build response
        :param requester_vk:
        :param request:
        :return:
        """
        requester_vk = request.sender

        assert self.store_full_blocks, "Must be able to store full blocks to reply to state update requests"
        self.log.debugv("Got block index request from sender {} requesting block hash {} my_vk {}"
                        .format(requester_vk, request.blockHash, self.verifying_key))

        if requester_vk == self.verifying_key:
            self.log.debugv("received request from myself dropping the req")
            return

        if self.is_caught_up:
            self.curr_hash = self.state.latest_block_hash
            self.curr_num = self.state.latest_block_num

        # tejas, latest_blk_num should correspond to request.block_hash or latest_num ?
        delta_idx = self.get_idx_list(vk=requester_vk,
                                      latest_blk_num=self.curr_num,
                                      sender_bhash=request.blockHash)

        if delta_idx is None:
            delta_idx = []

        self.log.debugv("Delta list {} for blk_num {} blk_hash {}".format(delta_idx, self.curr_num,
                                                                          request.blockHash))

        if delta_idx and len(delta_idx) > 1:
            assert delta_idx[0].get('blockNum') > delta_idx[-1].get('blockNum'), "ensure reply are in ascending order" \
                                                                                  " {}" .format(delta_idx)

        self._send_block_idx_reply(reply_to_vk=requester_vk,
                                   catchup_list=delta_idx)

    def _recv_blk_notif(self, update):
        # can get any time - hopefully one incremental request, how do you handle it in all cases?
        nw_blk_num = update.blockNum
        if self.is_caught_up:
            self.curr_hash = self.state.latest_block_hash
            self.curr_num = self.state.latest_block_num
            self.target_blk_num = self.curr_num
            self.awaited_blknum = self.curr_num
        if (nw_blk_num <= self.curr_num) or (nw_blk_num <= self.target_blk_num):
            return
        if nw_blk_num > (self.target_blk_num + 1):
            self.run_catchup()
        else:
            # actually you can request block data directly
            for vk in update.blockOwners:
                self.node_idx_reply_set.add(vk)
            self.is_caught_up = False
            self.target_blk_num = nw_blk_num
            if self.awaited_blknum == self.curr_num:
                self.awaited_blknum += 1
            for vk in update.blockOwners:
                self._send_block_data_req(mn_vk = vk, req_blk_num = nw_blk_num)

    def recv_new_blk_notif(self, update):
        self._recv_blk_notif(update)
        return self.is_catchup_done()

    # todo handle mismatch between redis and monodb
    # MASTER ONLY CALL
    def _send_block_idx_reply(self, reply_to_vk, catchup_list=[]):
        # this func doesnt care abt catchup_state we respond irrespective
        self.log.info("catchup list -> {}".format(catchup_list))

        msg_type, msg = Message.get_message_packed(
                                      MessageType.BLOCK_INDEX_REPLY,
                                      indices=catchup_list)

        self.log.debugv("Sending block index reply to vk {}, catchup {}".format(reply_to_vk, catchup_list))
        self.router.send_msg(filter=reply_to_vk,
                             msg_type=msg_type,
                             msg=msg)

    # MASTER ONLY CALL
    def recv_block_data_req(self, sender: bytes, req):
        block_num = req.blockNum
        blk_dict = self.driver.get_block(block_num)

        self.log.info(blk_dict)

        if '_id' in blk_dict:
            del blk_dict['_id']

        msg_type, msg = Message.get_message_packed(
                                      MessageType.BLOCK_DATA_REPLY,
                                      **blk_dict)

        self.router.send_msg(sender, msg=msg, msg_type=msg_type)

    def get_idx_list(self, vk, latest_blk_num, sender_bhash):
        # check if requester is master or del
        self.log.info(sender_bhash)
        valid_node = vk.decode() in PhoneBook.state_sync

        if valid_node:

            index = self.driver.get_index(sender_bhash)

            given_blk_num = index.get('blockNum') if index else 0

            self.log.debugv('given block is already latest hash - {} givenblk - {} curr-{}'
                            .format(sender_bhash, given_blk_num, latest_blk_num))

            if given_blk_num == latest_blk_num:
                self.log.debug('given block is already latest')
                return None
            else:
                idx_delta = self.driver.get_last_n(latest_blk_num - given_blk_num)
                return idx_delta

        assert valid_node, "invalid vk given key is not of master/delegate/statestync dumping vk {}".format(vk)

    # removed flooding, but it could be too sequential?
    # use futures to control rate of requests?
    def process_recv_idx(self):
        if (self.awaited_blknum <= self.curr_num) and (self.awaited_blknum < self.target_blk_num):
            self.awaited_blknum = self.curr_num + 1
            # don't request if it is in stashed list. move to next one
            while self.awaited_blknum in self.rcv_block_dict:
                self.awaited_blknum = self.awaited_blknum + 1
            blknum = 0
            blk_ptr = None
            while (blknum < self.awaited_blknum) and len(self.block_delta_list):
                blk_ptr = self.block_delta_list.pop(0)
                blknum = blk_ptr.get('blockNum')

            if blknum < self.awaited_blknum:
                return

            mn_list = blk_ptr.get('blockOwners')
            for vk in mn_list:
                self._send_block_data_req(mn_vk = vk, req_blk_num = self.awaited_blknum)

    def update_received_block(self, block=None):
        assert self.curr_num in self.rcv_block_dict, "not found the received block!"
        cur_num = self.curr_num
        while cur_num in self.rcv_block_dict:
            block = self.rcv_block_dict[cur_num]
            if self.store_full_blocks is True:
                update_blk_result = bool(self.driver.evaluate_wr(entry=block._data.to_dict()))
                assert update_blk_result is True, "failed to update block"

            self.state.update_with_block(block)
            self.curr_num = cur_num
            cur_num = cur_num + 1

        self.curr_hash = self.state.latest_block_hash
        self.curr_num = self.state.latest_block_num

    def _check_idx_reply_quorum(self):
        return len(self.node_idx_reply_set) >= self.my_quorum

    def is_catchup_done(self):
        if self.is_caught_up:
            return True
        self.is_caught_up = (self.target_blk_num == self.curr_num) and \
                            self._check_idx_reply_quorum()

        return self.is_caught_up





NETWORK_SIZE = '2-2-2'

MN_LOG_LVL = 11
WITNESS_LOG_LVL = 30
DELEGATE_LOG_LVL = 1# 11
SENECA_LOG_LVL = 11

# Set ENABLE_BAD_ACTORS to TRUE to enable periodic consensus failures
ENABLE_BAD_ACTORS = False
BAD_ACTOR_SET = {1}  # The indices of delegates who shall misbehave and periodically send bad SubBlockContender
BAD_SB_SET = {1}  # The indices of the sub-blocks to send bad SubBlockContenders for

RESET_DB = False


from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('{}.json'.format(NETWORK_SIZE))

from vmnet.testcase import BaseNetworkTestCase
import unittest, time, random, vmnet, cilantro_ee, os
from os.path import join, dirname
from cilantro_ee.utils.test.mp_test_case import vmnet_test
from cilantro_ee.utils.test.god import God
from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.utils.test.node_runner import *
from cilantro_ee.core.logger import get_logger, overwrite_logger_level
import logging, os, shutil, time
from cilantro_ee.constants.system_config import *


if os.getenv('USE_LOCAL_SENECA', '0') != '0':
    log = get_logger("SenecaCopier")
    user_sen_path = os.getenv('SENECA_PATH', None)
    assert user_sen_path, "If USE_LOCAL_SENECA env var is set, SENECA_PATH env var must also be set!"

    import cilantro_ee
    venv_sen_path = cilantro_ee.__path__[0] + '/../venv/lib/python3.6/site-packages/seneca'

    assert os.path.isdir(venv_sen_path), "Expect virtual env seneca at path {}".format(venv_sen_path)
    assert os.path.isdir(user_sen_path), "Expect user seneca at path {}".format(user_sen_path)

    log.debugv("Removing venv seneca at path {}".format(venv_sen_path))
    shutil.rmtree(venv_sen_path)

    log.notice("Copying user seneca path {} to venv packages at path {}".format(user_sen_path, venv_sen_path))
    shutil.copytree(user_sen_path, venv_sen_path)
    log.notice("Done copying")


class TestManualDump(BaseNetworkTestCase):

    NUM_BLOCKS = 2
    VOLUME = TRANSACTIONS_PER_SUB_BLOCK * NUM_SB_PER_BLOCK * NUM_BLOCKS  # Number of transactions to dump
    config_file = join(dirname(cilantro_ee.__path__[0]), 'vmnet_configs', 'cilantro_ee-{}-bootstrap.json'.format(NETWORK_SIZE))
    PROFILE_TYPE = None

    @vmnet_test(run_webui=True)
    def test_dump(self):
        log = get_logger("Dumpatron")
        cmd = 'docker container stop $(docker container ls -a -q -f name=delegate_8)'
        cmd1 = 'docker kill $(docker container ls -a -q -f name=delegate_8)'

        log.important("PORTS: {}".format(self.ports))

        # Bootstrap master
        for i, nodename in enumerate(self.groups['masternode']):
            self.execute_python(nodename, God.run_mn(i, log_lvl=MN_LOG_LVL, nonce_enabled=False, reset_db=RESET_DB),
                                async=True, profiling=self.PROFILE_TYPE)

        # Bootstrap witnesses
        for i, nodename in enumerate(self.groups['witness']):
            self.execute_python(nodename, God.run_witness(i, log_lvl=WITNESS_LOG_LVL, reset_db=RESET_DB), async=True,
                                profiling=self.PROFILE_TYPE)

        # Bootstrap delegates
        for i, nodename in enumerate(self.groups['delegate']):
            if ENABLE_BAD_ACTORS and i in BAD_ACTOR_SET:
                self.execute_python(nodename, God.run_delegate(i, log_lvl=DELEGATE_LOG_LVL, seneca_log_lvl=SENECA_LOG_LVL,
                                    reset_db=RESET_DB, bad_actor=True, bad_sb_set=BAD_SB_SET), async=True,
                                    profiling=self.PROFILE_TYPE)
            else:
                self.execute_python(nodename, God.run_delegate(i, log_lvl=DELEGATE_LOG_LVL, seneca_log_lvl=SENECA_LOG_LVL,
                                    reset_db=RESET_DB), async=True, profiling=self.PROFILE_TYPE)

        while True:
            user_input = input("Enter an integer representing the # of transactions to dump, or 'x' to quit.")

            if user_input.lower() == 's':
                log.important("stoping delegate 5")
                os.system(cmd)
                time.sleep(5)
                os.system(cmd1)

            if user_input.lower() == 'c':
                log.important("Testing catchup start delegate 8")
                self.execute_python('delegate_8', wrap_func(run_delegate, 3), async=True, profiling=self.PROFILE_TYPE)

            if user_input.lower() == 'x':
                log.important("Termination input detected. Breaking")
                break

            vol = int(user_input) if user_input.isdigit() else self.VOLUME
            log.important3("Dumpatron dumping {} transactions!".format(vol))
            self.execute_python('mgmt', God.dump_it(volume=vol), async=True, profiling=self.PROFILE_TYPE)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

__version__ = '0.0.1'

with open("README.md", "r") as fh:
    long_desc = fh.read()

setup(
    name='cilantro_ee',
    version=__version__,
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'Cython==0.29',
        'pycapnp==0.6.3',
        #PyNaCl==1.2.1
        'pyzmq==17.0.0',
        'requests==2.20.0',
        'uvloop==0.9.1',
        'u-msgpack-python==2.5.0',
        'yarl==1.1.0',
        'contracting',
        'click',
        'simple-crypt',
        'sanic==19.6.3',
        'pymongo==3.7.2'
    ],
    extras_require={
        'dev': open('dev-requirements.txt').readlines()
    },
    entry_points={
        'console_scripts': [
            'storage=cilantro_ee.networking.storage:serve',
            'witness=cilantro_ee.networking.witness:serve',
            'cil=cilantro_ee.cli:main'
        ],
    },
    zip_safe=False,
    package_data={
        '': [],
        'cilantro_ee': ['cilantro_ee.conf'],
    },
    description = "Lamden Blockchain",
    long_description= long_desc,
    long_description_content_type="text/markdown",
    url='https://github.com/Lamden/cilantro-enterprise',
    author='Lamden',
    author_email='team@lamden.io',
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6.5',
)



from cilantro_ee.core.utils.worker import Worker

import os
import asyncio
import zmq.asyncio
import time


PORT = 9432
PORT_INSECURE = 9433
PROTOCOL = 'tcp'

PUB_SOCK_KEY = 'default_pub'
SUB_SOCK_KEY = 'default_sub'
DEFAULT_DOMAIN = '*'


def wrap_func(fn, *args, **kwargs):
    def wrapper():
        return fn(*args, **kwargs)
    return wrapper


class PubSubAuthTester(Worker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log.important3("PubSubAuthTester starting with VK {}".format(self.verifying_key))

        # Dict of socket key -> LSocket instance
        self.pub_sockets = {}
        self.sub_sockets = {}

    def start(self):
        self.loop.run_forever()

    def add_pub_socket(self, ip, port=PORT, protocol=PROTOCOL, secure=False, domain=DEFAULT_DOMAIN, socket_key=PUB_SOCK_KEY):
        assert socket_key not in self.pub_sockets, "Key {} already exists in pub sockets {}".format(socket_key, self.pub_sockets)

        sock = self.manager.create_socket(zmq.PUB, secure=secure, domain=domain)
        self.log.socket("Binding pub socket with key {} using ip {}".format(socket_key, ip))
        sock.bind(port=PORT, protocol=protocol, ip=ip)

        self.pub_sockets[socket_key] = sock

    def add_sub_socket(self, filter=b'', secure=False, domain=DEFAULT_DOMAIN, socket_key=SUB_SOCK_KEY):
        assert socket_key not in self.sub_sockets, "Key {} already exists in sub sockets {}".format(socket_key, self.sub_sockets)
        assert type(filter) is bytes, "Filter must be bytes, not {}".format(filter)

        sock = self.manager.create_socket(zmq.SUB, secure=secure, domain=domain)
        sock.setsockopt(zmq.SUBSCRIBE, filter)
        sock.add_handler(handler_func=self.handle_sub, start_listening=True)

        self.sub_sockets[socket_key] = sock

    def connect_sub(self, socket_key=SUB_SOCK_KEY, vk='', ip='', port=PORT, protocol=PROTOCOL):
        assert socket_key in self.sub_sockets, "Key {} not found in sub sockets {}".format(socket_key, self.sub_sockets)

        self.log.important("connecting to vk {}".format(vk))

        sock = self.sub_sockets[socket_key]
        sock.connect(port=port, protocol=protocol, vk=vk, ip=ip)
        time.sleep(2) # TODO for davis

    def start_publishing(self, num_msgs=50, interval=1, filter=b''):
        assert len(self.pub_sockets) > 0, "Must add at least 1 pub socket to start publishing"
        async def _start_pubbing(num):
            for i in range(num):
                self.log.info("sending pub {}".format(i))
                msg = [filter, 'hello #{} from {}'.format(i, os.getenv('HOST_IP')).encode()]
                for sock in self.pub_sockets.values():
                    sock.send_multipart(msg)
                await asyncio.sleep(interval)

        asyncio.ensure_future(_start_pubbing(num_msgs))

    def send_pub(self, msg: bytes, filter: bytes=b'', socket_key=PUB_SOCK_KEY):
        assert socket_key in self.pub_sockets, "Key {} not found in pub sockets {}".format(socket_key, self.pub_sockets)
        assert type(msg) is bytes, "Msg arg must be bytes, not {}".format(msg)

        self.log.debugv("Sending msg {} with filter {} on socket with key {}".format(msg, filter, socket_key))
        self.pub_sockets[socket_key].send_multipart([filter, msg])

    def handle_sub(self, frames):
        self.log.important("got <secure> pub msg with frames {}".format(frames))



from cilantro_ee.messages.transaction.data import TransactionData
from cilantro_ee.utils import Hasher
from cilantro_ee.messages.transaction.contract import ContractTransactionBuilder, ContractTransaction
from cilantro_ee.messages.block_data.block_data import BlockData, GENESIS_BLOCK_HASH, BlockDataBuilder
from cilantro_ee.messages.block_data.state_update import BlockDataReply
from cilantro_ee.messages.block_data.sub_block import SubBlock, SubBlockBuilder
from cilantro_ee.core.containers.merkle_tree import MerkleTree
from unittest import TestCase
import unittest
from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
TEST_SK, TEST_VK = TESTNET_MASTERNODES[0]['sk'], TESTNET_MASTERNODES[0]['vk']


class TestBlockData(TestCase):

    def test_create(self):
        input_hash1 = 'A'*64
        input_hash2 = 'B'*64
        sb1 = SubBlockBuilder.create(input_hash=input_hash1, idx=0)
        sb2 = SubBlockBuilder.create(input_hash=input_hash2, idx=1)
        sbs = [sb1, sb2]

        prev_b_hash = GENESIS_BLOCK_HASH
        block_hash = BlockData.compute_block_hash([sb1.merkle_root, sb2.merkle_root], prev_b_hash)
        block_num = 1
        block_owners = [TEST_VK]

        block = BlockData.create(block_hash=block_hash, prev_block_hash=prev_b_hash, block_num=block_num,
                                 sub_blocks=sbs, block_owners=block_owners)

        self.assertEqual(block.block_hash, block_hash)
        self.assertEqual(block.block_num, block_num)
        self.assertEqual(block.block_owners, block_owners)
        self.assertEqual(block.prev_block_hash, prev_b_hash)

        self.assertEqual(block.transactions, sb1.transactions + sb2.transactions)
        self.assertEqual(block.merkle_roots, [sb1.merkle_root, sb2.merkle_root])
        self.assertEqual(block.input_hashes, [sb1.input_hash, sb2.input_hash])
        self.assertEqual(block.merkle_leaves, sb1.merkle_leaves + sb2.merkle_leaves)

    def test_serialize_deserialize(self):
        input_hash1 = 'A'*64
        input_hash2 = 'B'*64
        sb1 = SubBlockBuilder.create(input_hash=input_hash1, idx=0)
        sb2 = SubBlockBuilder.create(input_hash=input_hash2, idx=1)
        sbs = [sb1, sb2]

        prev_b_hash = GENESIS_BLOCK_HASH
        block_hash = BlockData.compute_block_hash([sb1.merkle_root, sb2.merkle_root], prev_b_hash)
        block_num = 1
        block_owners = [TEST_VK]

        block = BlockData.create(block_hash=block_hash, prev_block_hash=prev_b_hash, block_num=block_num,
                                 sub_blocks=sbs, block_owners=block_owners)

        clone = BlockData.from_bytes(block.serialize())
        self.assertEqual(clone, block)

    def test_get_tx_hash_to_merkle_leaf(self):
        input_hash1 = 'A'*64
        input_hash2 = 'B'*64
        sb1 = SubBlockBuilder.create(input_hash=input_hash1, idx=0)
        sb2 = SubBlockBuilder.create(input_hash=input_hash2, idx=1)
        sbs = [sb1, sb2]

        prev_b_hash = GENESIS_BLOCK_HASH
        block_hash = BlockData.compute_block_hash([sb1.merkle_root, sb2.merkle_root], prev_b_hash)
        block_num = 1
        block_owners = [TEST_VK]

        block = BlockData.create(block_hash=block_hash, prev_block_hash=prev_b_hash, block_num=block_num,
                                 sub_blocks=sbs, block_owners=block_owners)

        tx_hash_to_leaves = block.get_tx_hash_to_merkle_leaf()

        # Make sure lengths match up
        self.assertEqual(len(block.merkle_leaves), len(tx_hash_to_leaves))

    def test_block_data_builder(self):
        data = BlockDataBuilder.create_random_block(prev_hash=GENESIS_BLOCK_HASH, num=1)
        self.assertEqual(data.prev_block_hash, GENESIS_BLOCK_HASH)
        self.assertEqual(data.block_num, 1)

    def test_block_data_builder_conseq_blocks(self):
        num = 5
        blocks = BlockDataBuilder.create_conseq_blocks(num)
        curr_hash, curr_num = GENESIS_BLOCK_HASH, 0
        for block in blocks:
            self.assertEqual(block.prev_block_hash, curr_hash)
            self.assertEqual(block.block_num, curr_num+1)
            curr_hash = block.block_hash
            curr_num += 1

    def test_block_data_reply_create_from_block(self):
        block = BlockDataBuilder.create_random_block()
        bd_reply = BlockDataReply.create_from_block(block)

        self.assertEqual(block.block_hash, bd_reply.block_hash)
        self.assertEqual(block.block_num, bd_reply.block_num)
        self.assertEqual(block.block_owners, bd_reply.block_owners)
        self.assertEqual(block.prev_block_hash, bd_reply.prev_block_hash)
        self.assertEqual(block.transactions, bd_reply.transactions)

    def test_from_dict(self):
        block = BlockDataBuilder.create_random_block()
        clone_from_bytes = BlockData.from_bytes(block.serialize())
        clone_from_dict = BlockData.from_dict(block.to_dict())
        clone_from_bytes_to_dict = BlockData.from_dict(clone_from_bytes.to_dict())

        self.assertEqual(block, clone_from_bytes)
        self.assertEqual(clone_from_bytes, clone_from_dict)
        self.assertEqual(clone_from_dict, clone_from_bytes_to_dict)


if __name__ == '__main__':
    unittest.main()



from multiprocessing import Queue
from cilantro_ee.utils.lprocess import LProcess

from cilantro_ee.nodes.base import NodeBase
from cilantro_ee.nodes.masternode.transaction_batcher import TransactionBatcher
from cilantro_ee.nodes.masternode.block_aggregator import BlockAggregator
from cilantro_ee.nodes.masternode.webserver import start_webserver

import os, random


IPC_IP = 'masternode-ipc-sock'
IPC_PORT = 6967


class Masternode(NodeBase):

    # This call should not block!
    def start_node(self):
        self.tx_queue = Queue()
        self.ipc_ip = IPC_IP + '-' + str(os.getpid()) + '-' + str(random.randint(0, 2**32))

        self._start_web_server()
        if not os.getenv('MN_MOCK'):  # TODO @stu do we need this still? --davis
            self._start_batcher()
            self._start_block_agg()
            return 1
        else:
            self.log.warning("MN_MOCK env var is set! Not starting block aggregator or tx batcher.")
            return 0

    def _start_web_server(self):
        self.log.debug("Masternode starting REST server on port 8080")
        self.server = LProcess(target=start_webserver, name='WebServerProc', args=(self.tx_queue,))
        self.server.start()

    def _start_batcher(self):
        # Create a worker to do transaction batching
        self.log.info("Masternode starting transaction batcher process")
        self.batcher = LProcess(target=TransactionBatcher, name='TxBatcherProc',
                                kwargs={'queue': self.tx_queue, 'signing_key': self.signing_key,
                                        'ip': self.ip, 'ipc_ip': self.ipc_ip, 'ipc_port': IPC_PORT})
        self.batcher.start()

    def _start_block_agg(self):
        self.log.info("Masternode starting BlockAggregator Process")
        self.block_agg = LProcess(target=BlockAggregator,  name='BlockAgg', kwargs={'ip': self.ip, 'ipc_ip': self.ipc_ip, 'ipc_port': IPC_PORT, 'signing_key': self.signing_key, 'name': 'BlockAgg'})
        self.block_agg.start()



from unittest import TestCase
from unittest.mock import patch, MagicMock

from cilantro_ee.core.containers.merkle_tree import MerkleTree
from cilantro_ee.constants.system_config import *
from cilantro_ee.nodes.masternode.block_contender import BlockContender


# TODO write these
class TestBlockContender(TestCase):
    pass

